
' TODO
' IntelliSense
' http://www.actiprosoftware.com/products/dotnet/WindowsForms/SyntaxEditor/Default.aspx

' Set proper casing for common keywords
' For / If / End If / While etc...

' WM_MOUSEWHEEL lost events
' http://www.cplusplus.com/forum/windows/33960/


Imports System.CodeDom.Compiler
Imports System.Text.RegularExpressions
Imports System.Reflection


' ===============================================================================================================
'   RichTextBox overloaded to MyRichTextBox
' ===============================================================================================================
Public Class MyRichTextBox
    Inherits RichTextBox
    '
    Private Const WM_MOUSEWHEEL As Int32 = 522
    Private Const EM_SCROLL As Int32 = &HB5
    Private Const WM_VSCROLL As Int32 = &H115
    Private Const SB_LINEUP As Int32 = 0
    Private Const SB_LINEDOWN As Int32 = 1
    '
    Protected Overrides Sub WndProc(ByRef m As Message)
        If m.Msg = WM_MOUSEWHEEL Then
            ' ------------------------------------------------------------- scroll suggestion list
            If LV_suggestions.Visible Then
                RTB_ScrollSuggestionList(m.WParam.ToInt32 \ &H10000)
                Exit Sub
            End If
            ' ------------------------------------------------------------- scroll the main script window
            m.Msg = EM_SCROLL ' WM_VSCROLL 'EM_SCROLL
            m.LParam = IntPtr.Zero
            If m.WParam.ToInt32 > 0 Then
                m.WParam = CType(SB_LINEUP, IntPtr)
            Else
                m.WParam = CType(SB_LINEDOWN, IntPtr)
            End If
            ' ------------------------------------------------------------- with variable speed
            Static oldtime As DateTime
            Dim delta As Double = 0.05 / DateTime.Now.Subtract(oldtime).TotalSeconds
            oldtime = DateTime.Now
            delta = delta ^ 2
            delta = Math.Max(delta, 3)
            delta = Math.Min(delta, 8)
            For i As Int32 = 1 To CInt(delta)
                MyBase.WndProc(m)
            Next
            Exit Sub
        End If
        MyBase.WndProc(m)
    End Sub
End Class


' ===============================================================================================================
'   EDITOR
' ===============================================================================================================
Module Module_Editor

    Friend Editor_RunningCode As Boolean
    Friend Editor_Modified As Boolean = False
    Friend Editor_CommentChars As String = "'"

    Private EventsAreEnabled As Boolean = False

    Private FontName As String
    Private TabSize As Int32
    Private Zoom As Single = 1

    Private Color_NormalText As Color = Color.Black
    Private Color_QuotedText As Color = Color.DarkRed
    Private Color_KeywordText As Color = Color.Blue
    Private Color_CommentText As Color = Color.Green
    Private Color_ErrorBack As Color = Color.FromArgb(255, 220, 210)
    Private Color_WarningBack As Color = Color.FromArgb(220, 255, 210)
    Private Color_TextBack As Color = Color.FromArgb(245, 245, 245)
    Private Color_Back As Color = Color.FromArgb(255, 255, 255)

    Private WithEvents RichTextBoxCode As MyRichTextBox = Nothing
    Private WithEvents ListViewErrors As ListView = Nothing
    Private WithEvents TextBoxRunTimeErrors As TextBox = Nothing
    Private WithEvents StatusLabel As ToolStripStatusLabel
    Private TabReports As TabControl = Nothing

    Friend WithEvents LV_suggestions As ListView
    Private Editor_Form As Form


    Friend Sub Editor_Init(ByVal _EditorForm As Form, _
                           ByRef _RichTextBoxCode As MyRichTextBox, _
                           Optional ByRef _ListViewErrors As ListView = Nothing, _
                           Optional ByRef _TabReports As TabControl = Nothing, _
                           Optional ByRef _TextBoxRunTimeErrors As TextBox = Nothing, _
                           Optional ByRef _StatusLabel As ToolStripStatusLabel = Nothing, _
                           Optional ByRef ListView_Suggestion_Icons As ImageList = Nothing)

        Editor_Form = _EditorForm
        RichTextBoxCode = _RichTextBoxCode
        TabReports = _TabReports
        ListViewErrors = _ListViewErrors
        TextBoxRunTimeErrors = _TextBoxRunTimeErrors
        StatusLabel = _StatusLabel

        ' -------------------------------------------------------------------- Editor_Form
        LV_suggestions = New ListView
        Editor_Form.Controls.Add(LV_suggestions)
        With LV_suggestions
            .Sorting = SortOrder.Ascending
            .BorderStyle = BorderStyle.FixedSingle
            .FullRowSelect = True
            .HeaderStyle = ColumnHeaderStyle.None
            .View = View.Details
            .MultiSelect = False
            .Columns.Add("")
            .SmallImageList = ListView_Suggestion_Icons
            .BackColor = Color.FromArgb(255, 250, 240)
            .Visible = False
            .HideSelection = False
        End With

        ' -------------------------------------------------------------------- RichTextBoxCode
        With RichTextBoxCode
            .AcceptsTab = True
            .DetectUrls = False
            .WordWrap = False
            .HideSelection = False
            .ShowSelectionMargin = True  ' if true then scrollbars are disabled at start
        End With

        ' -------------------------------------------------------------------- ListViewErrors
        If ListViewErrors IsNot Nothing Then
            AddHandler ListViewErrors.SelectedIndexChanged, AddressOf ListViewErrors_SelectedIndexChanged
        End If

        ' -------------------------------------------------------------------- StatusLabel
        If StatusLabel IsNot Nothing Then
            StatusLabel.DoubleClickEnabled = True
            AddHandler StatusLabel.DoubleClick, AddressOf StatusLabel_DoubleClick
        End If

    End Sub

    Friend Sub Editor_SetFont(ByVal font As String)
        RTB_SuspendRedraw()
        '
        If font = "Courier New" Then
            RichTextBoxCode.Font = New Font("Courier New", 10)
        Else
            RichTextBoxCode.Font = New Font("Lucida Console", 10, FontStyle.Regular)
        End If
        '
        RTB_ResumeRedraw()
    End Sub

    Friend Sub Editor_SetZoom(ByVal ZoomPercent As Int32)
        Dim z As Single = CSng(ZoomPercent / 100)
        If z > 1.44 Then
            z = 1.5
        ElseIf z > 1.31 Then
            z = 1.38
        ElseIf z > 1.18 Then
            z = 1.25
        ElseIf z > 1.06 Then
            z = 1.12
        ElseIf z > 0.94 Then
            z = 1
        ElseIf z > 0.81 Then
            z = 0.88
        ElseIf z > 0.68 Then
            z = 0.75
        ElseIf z > 0.56 Then
            z = 0.62
        Else
            z = 0.5
        End If
        RichTextBoxCode.ZoomFactor = z
        'StatusLabel.Text = z.ToString
        'StatusLabel.Text = RichTextBoxCode.ZoomFactor.ToString
        'StatusLabel.Text = RichTextBoxCode.Font.Unit.ToString
    End Sub

    Friend Sub Editor_SetTabSize(ByVal _tabSize As String)
        RTB_SuspendRedraw()
        '
        RichTextBoxCode.SelectAll()
        TabSize = CInt(Val(_tabSize))
        RTB_SetTabPositions()
        '
        RTB_ResumeRedraw()
    End Sub

    Friend Sub Editor_SetColors(ByVal NormalText As Color, _
                                ByVal QuotedText As Color, _
                                ByVal KeywordText As Color, _
                                ByVal CommentText As Color, _
                                ByVal ErrorBack As Color, _
                                ByVal WarningBack As Color, _
                                ByVal TextBack As Color, _
                                ByVal Back As Color)

        Color_NormalText = NormalText
        Color_QuotedText = QuotedText
        Color_KeywordText = KeywordText
        Color_CommentText = CommentText
        Color_ErrorBack = ErrorBack
        Color_WarningBack = WarningBack
        Color_TextBack = TextBack
        Color_Back = Back
    End Sub

    Friend Sub Editor_SetCursor_Wait()
        Application.DoEvents()
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor
    End Sub

    Friend Sub Editor_SetCursor_Normal()
        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
    End Sub

    Friend Function Editor_IsTextselected() As Boolean
        Return RichTextBoxCode.SelectedText.Length > 0
    End Function

    Friend Sub Editor_RunningState_Set()
        Editor_RunningCode = True
        RichTextBoxCode.Enabled = False
    End Sub

    Friend Sub Editor_RunningState_Reset()
        RichTextBoxCode.Enabled = True
        Editor_RunningCode = False
    End Sub

    Friend Sub Editor_Indent()
        RTB_IndentSelectedLines(1)
    End Sub
    Friend Sub Editor_UnIndent()
        RTB_IndentSelectedLines(-1)
    End Sub
    Friend Sub Editor_Comment()
        RTB_CommentSelectedLines(True)
    End Sub
    Friend Sub Editor_UnComment()
        RTB_CommentSelectedLines(False)
    End Sub


    ' ===============================================================================================================
    '  OverWriteMode / Caret / Status 
    ' ===============================================================================================================
    Private Editor_OverwriteMode As Boolean = False

    Private Sub RichTextBoxCode_PreviewKeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.PreviewKeyDownEventArgs) Handles RichTextBoxCode.PreviewKeyDown
        Select Case e.KeyCode
            Case Keys.Insert
                Editor_OverwriteMode = Not Editor_OverwriteMode
                ShowCaret()
                UpdateStatusStrip()
            Case Keys.CapsLock, Keys.NumLock, Keys.Scroll
                UpdateStatusStrip()
        End Select
    End Sub

    Public Declare Auto Sub CreateCaret Lib "user32" (ByVal hWnd As IntPtr, ByVal hBitmap As Integer, ByVal nWidth As Integer, ByVal nHeight As Integer)
    Public Declare Auto Sub ShowCaret Lib "user32" (ByVal hWnd As IntPtr)
    Private Sub ShowCaret()
        If OperatingSystemIsWindows Then
            Dim x As Int32 = CInt(8 * RichTextBoxCode.ZoomFactor)
            Dim y As Int32 = CInt(13 * RichTextBoxCode.ZoomFactor)
            If Editor_OverwriteMode Then
                CreateCaret(RichTextBoxCode.Handle, 0, x, y)
                ShowCaret(RichTextBoxCode.Handle)
            Else
                CreateCaret(RichTextBoxCode.Handle, 0, 1, y)
                ShowCaret(RichTextBoxCode.Handle)
            End If
        End If
    End Sub



    ' -----------------------------------------------------------------------------------------------------
    '  STATUS STRIP
    ' -----------------------------------------------------------------------------------------------------
    Private Sub StatusLabel_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles StatusLabel.DoubleClick
        Editor_Goto()
    End Sub

    Private Sub UpdateStatusStrip()
        If StatusLabel Is Nothing Then Exit Sub
        Dim s As String = ""
        s &= "Ln " & RTB_GetFirstSelectedLine() + 1
        s &= "        Col " & RTB_GetCursorColumn()
        If Editor_OverwriteMode Then s &= "        OVERWRITE  " Else s &= "        INSERT  "
        If My.Computer.Keyboard.CapsLock Then s &= " CAPS-LOCK"
        If My.Computer.Keyboard.NumLock Then s &= " NUM-LOCK"
        If My.Computer.Keyboard.ScrollLock Then s &= " SCROLL-LOCK"
        StatusLabel.Text = s
    End Sub

    Private Function RTB_GetCursorColumn() As Int32
        Dim p As Point = RichTextBoxCode.GetPositionFromCharIndex(RichTextBoxCode.SelectionStart)
        Return CInt(p.X / 8 / RichTextBoxCode.ZoomFactor)
    End Function

    Friend Sub Editor_Goto()
        Dim response As String = InputBox(vbCr & vbCr & vbCr & "Line number ?  ( 1 to " & RichTextBoxCode.Lines.Count & ")", _
                                        "Go To Line", _
                                        (RTB_GetFirstSelectedLine() + 1).ToString)

        Dim line As Int32 = CInt(Val(response)) - 1
        If line >= 0 And line < RichTextBoxCode.Lines.Count Then
            Dim lineStart As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(line)
            Dim lineLength As Int32 = RichTextBoxCode.Lines(line).Length
            RichTextBoxCode.Select(lineStart, lineLength)
        End If
    End Sub


    ' -----------------------------------------------------------------------------------------------------
    '  Select complete words also with underscores 
    ' -----------------------------------------------------------------------------------------------------
    Private Sub RTB_SelectCompleteWord()
        Dim ww As WholeWord = FindWholeWord(RichTextBoxCode.Text, RichTextBoxCode.SelectionStart)
        If ww.Len > 0 Then RichTextBoxCode.Select(ww.Start, ww.Len)
    End Sub


    ' -----------------------------------------------------------------------------------------------------
    '  Find WholeWord
    ' -----------------------------------------------------------------------------------------------------
    Private Structure WholeWord
        Dim Word As String
        Dim Start As Int32
        Dim Len As Int32
    End Structure
    Private Function FindWholeWord(ByVal text As String, ByVal index As Int32) As WholeWord
        Dim FirstCh As Int32 = index
        Dim LastCh As Int32 = index
        ' ----------------------------------------------------------------- search backward
        Do
            If FirstCh < 1 Then Exit Do
            FirstCh -= 1
            If Not Char.IsLetterOrDigit(text(FirstCh)) And Not text(FirstCh) = "_" Then
                FirstCh += 1
                Exit Do
            End If
        Loop
        ' ----------------------------------------------------------------- search forward
        Do
            If LastCh > text.Count - 1 Then LastCh -= 1 : Exit Do
            If Not Char.IsLetterOrDigit(text(LastCh)) And Not text(LastCh) = "_" Then
                LastCh -= 1
                Exit Do
            End If
            LastCh += 1
        Loop
        ' ------------------------------------------------------------------ set WholeWord values
        FindWholeWord.Start = FirstCh
        FindWholeWord.Len = LastCh - FirstCh + 1
        FindWholeWord.Word = text.Substring(FindWholeWord.Start, FindWholeWord.Len)
    End Function


    ' -----------------------------------------------------------------------------------------------------
    '  return true if "index" is not after a commented char 
    ' -----------------------------------------------------------------------------------------------------
    Private Function IsCommented(ByVal text As String, ByVal index As Int32) As Boolean
        Dim i As Int32 = InStr(text, Editor_CommentChars)
        If i > 0 And i <= index Then
            Return True
        Else
            Return False
        End If
    End Function

    ' -----------------------------------------------------------------------------------------------------
    '  return true if Char "CharIndex" is not after a commented char 
    ' -----------------------------------------------------------------------------------------------------
    Private Function IsCommented(ByVal CharIndex As Int32) As Boolean
        Dim Line As Int32
        Dim CharIndexInLine As Int32
        Line = RichTextBoxCode.GetLineFromCharIndex(CharIndex)
        CharIndexInLine = CharIndex - RichTextBoxCode.GetFirstCharIndexFromLine(Line)
        Return IsCommented(RichTextBoxCode.Lines(Line), CharIndexInLine)
    End Function


    ' -----------------------------------------------------------------------------------------------------
    '  Show OpenClose words   ( not working because it forces the cursor at the start of each line )
    ' -----------------------------------------------------------------------------------------------------
    'Private Sub OpenClose_Test()
    '    If My.Computer.Keyboard.ShiftKeyDown Or Control.MouseButtons = MouseButtons.Left Then Exit Sub

    '    Dim i As Int32 = RichTextBoxCode.SelectionStart
    '    FindWholeWord(i)

    '    ' ----------------------------------------------------------- WholeWord TESTER
    '    OpenClose_UnMark()
    '    OpenClose_Mark(WholeWord_Start, WholeWord_Len)


    '    'Dim s As String = WholeWord.ToLower
    '    'Select Case s
    '    '    Case "("
    '    '        OpenClose_Mark(i, 1)
    '    '    Case ")"
    '    '        OpenClose_Mark(i, 1)
    '    '    Case "sub", "function", "if", "else", "end"
    '    '        OpenClose_Mark(WholeWord_Start, WholeWord_Len)
    '    '    Case Else
    '    '        If OpenClose_Marked Then OpenClose_UnMark()
    '    'End Select
    'End Sub

    'Structure OpenClose_Positions
    '    Dim start1 As Int32
    '    Dim len1 As Int32
    '    Dim start2 As Int32
    '    Dim len2 As Int32
    'End Structure
    'Private OpenClose_Old As OpenClose_Positions
    'Private OpenClose_New As OpenClose_Positions
    'Private OpenClose_Marked As Boolean = False

    'Private Sub OpenClose_Mark(ByVal start As Int32, ByVal len As Int32)

    '    OpenClose_New.start1 = start
    '    OpenClose_New.len1 = len

    '    ' ----------------------------------------------------------------------- save old situation
    '    EventsAreEnabled = False
    '    Undo_EnableAdd = False
    '    RTB_SuspendRedraw()
    '    RTB_SaveSelectionAndScroll()
    '    ' ----------------------------------------------------------------------- remove old marks
    '    RichTextBoxCode.Select(OpenClose_Old.start1, OpenClose_Old.len1)
    '    RichTextBoxCode.SelectionBackColor = Color_Back

    '    ' ----------------------------------------------------------------------- show new marks
    '    RichTextBoxCode.Select(OpenClose_New.start1, OpenClose_New.len1)
    '    RichTextBoxCode.SelectionBackColor = Color.YellowGreen

    '    ' ----------------------------------------------------------------------- old = new
    '    OpenClose_Old = OpenClose_New
    '    OpenClose_Marked = True

    '    ' ----------------------------------------------------------------------- restore old situation
    '    RTB_RestoreSelectionAndScroll()
    '    RTB_ResumeRedraw()
    '    Undo_EnableAdd = True
    '    EventsAreEnabled = True
    'End Sub

    'Private Sub OpenClose_UnMark()
    '    ' ----------------------------------------------------------------------- save old situation
    '    EventsAreEnabled = False
    '    Undo_EnableAdd = False
    '    RTB_SuspendRedraw()
    '    RTB_SaveSelectionAndScroll()
    '    ' ----------------------------------------------------------------------- remove old marks
    '    RichTextBoxCode.Select(OpenClose_Old.start1, OpenClose_Old.len1)
    '    RichTextBoxCode.SelectionBackColor = Color_Back
    '    ' ----------------------------------------------------------------------- restore old situation
    '    RTB_RestoreSelectionAndScroll()
    '    RTB_ResumeRedraw()

    '    Undo_EnableAdd = True
    '    EventsAreEnabled = True
    '    '
    '    OpenClose_Marked = False
    'End Sub

    ' -----------------------------------------------------------------------------------------------------
    '  MOUSE MOVE TIMER
    ' -----------------------------------------------------------------------------------------------------
    Private Sub RestartMouseMoveTimer()
        Timer2.Stop()
        Timer2.Interval = 150
        Timer2.Start()
    End Sub
    Private WithEvents Timer2 As System.Windows.Forms.Timer = New System.Windows.Forms.Timer
    Private Sub Timer2_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Timer2.Tick
        Timer2.Stop()
        Suggestions_FindInfo()
    End Sub

    ' -----------------------------------------------------------------------------------------------------
    '  TEXT CHANGED TIMER
    ' -----------------------------------------------------------------------------------------------------
    Private Sub RestartTextChangedTimer()
        Timer1.Stop()
        Timer1.Interval = 300
        Timer1.Start()
    End Sub
    Private WithEvents Timer1 As System.Windows.Forms.Timer = New System.Windows.Forms.Timer
    Private Sub Timer1_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Timer1.Tick
        Timer1.Stop()
        Editor_UpdateAndCompile()
        Suggestions_Editing = True
    End Sub


    ' ===============================================================================================================
    '  EVENTS
    ' ===============================================================================================================
    Private Sub RichTextBoxCode_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles RichTextBoxCode.MouseMove
        If Not EventsAreEnabled Then Exit Sub
        RestartMouseMoveTimer()
    End Sub
    Private Sub RichTextBoxCode_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles RichTextBoxCode.TextChanged
        If Not EventsAreEnabled Then Exit Sub
        Undo_AddEntry()
        RestartTextChangedTimer()
    End Sub
    Private Sub RichTextBoxCode_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles RichTextBoxCode.MouseDown
        Dim chIndex As Int32 = RichTextBoxCode.GetCharIndexFromPosition(e.Location)
        '
        If e.Button = MouseButtons.Left Or chIndex < RichTextBoxCode.SelectionStart Or _
                                           chIndex > RichTextBoxCode.SelectionStart + RichTextBoxCode.SelectionLength Then
            ' ------------------------------------------------- do nothing if it is the last caracter
            If chIndex < RichTextBoxCode.Text.Count - 1 Then
                RichTextBoxCode.Select(chIndex, 0)
            End If
        End If
    End Sub
    Private Sub RichTextBoxCode_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles RichTextBoxCode.Click
        ShowCaret()
        Suggestions_CloseAll()
    End Sub
    Private Sub RichTextBoxCode_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles RichTextBoxCode.DoubleClick
        RTB_SelectCompleteWord()
        Suggestions_FindMembers()
    End Sub
    Private Sub RichTextBoxCode_HScroll(ByVal sender As Object, ByVal e As System.EventArgs) Handles RichTextBoxCode.HScroll
        ShowCaret()
    End Sub
    Private Sub RichTextBoxCode_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles RichTextBoxCode.SelectionChanged
        If Not EventsAreEnabled Then Exit Sub
        'OpenClose_Test()
        ShowCaret()
        UpdateStatusStrip()
    End Sub
    Private Sub RichTextBoxCode_VScroll(ByVal sender As Object, ByVal e As System.EventArgs) Handles RichTextBoxCode.VScroll
        ShowCaret()
        LV_suggestions_HideToolTip()
    End Sub


    ' ---------------------------------------------------------------------------- SCROLL SUGGESTIONS LIST VIEW
    Friend Sub RTB_ScrollSuggestionList(ByVal WheelDelta As Int32)
        Static oldtime As DateTime
        Dim delta As Double = 0.05 / DateTime.Now.Subtract(oldtime).TotalSeconds
        oldtime = DateTime.Now
        delta = delta ^ 2
        delta = Math.Max(delta, 1)
        delta = Math.Min(delta, 30)
        'Debug.Print(delta.ToString("0.0"))
        ListView_ChangeSelection_Relative(LV_suggestions, CInt(delta) * -Math.Sign(WheelDelta))
        Suggestions_ShowListViewToolTip()
    End Sub


    ' --------------------------------------- Removed selecting with SPACE to avoid to selecting unwanted terms
    ' --------------------------------------- Removed selecting with ENTER to avoid to selecting unwanted terms
    Private Const SelectWithSpace As Boolean = False
    Private Const SelectWithEnter As Boolean = False

    ' -----------------------------------------------------------------------------------------------------
    '  KEY DOWN - KEY UP
    ' -----------------------------------------------------------------------------------------------------
    Private Sub RichTextBoxCode_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles RichTextBoxCode.KeyDown
        '
        Select Case e.KeyCode

            Case Keys.Enter
                ' ------------------------------------------------------------ exit if selecting a suggestion
                If SelectWithEnter Then
                    If LV_suggestions.Visible Then Exit Select
                Else
                    LV_suggestions.Visible = False
                End If
                ' ------------------------------------------------------------ INDENTED NEW LINE
                If RichTextBoxCode.TextLength > 0 Then
                    Dim CaretPosition As Int32 = RichTextBoxCode.SelectionStart
                    Dim CaretLine As Int32 = RichTextBoxCode.GetLineFromCharIndex(CaretPosition)
                    Dim nTabs As Int32 = STR_CountInitialTabs(RichTextBoxCode.Lines(CaretLine), TabSize)
                    RichTextBoxCode.SelectedText = vbLf & StrDup(nTabs, vbTab)
                    RichTextBoxCode.Select(CaretPosition + 1 + nTabs, 0)
                    e.SuppressKeyPress = True
                End If

            Case Keys.Tab
                ' ------------------------------------------------------------ TAB and SHIT-TAB
                If RTB_GetFirstSelectedLine() <> RTB_GetLastSelectedLine() Then
                    If e.Shift Then
                        RTB_IndentSelectedLines(-1)
                    Else
                        RTB_IndentSelectedLines(1)
                    End If
                    e.SuppressKeyPress = True
                End If

            Case Keys.V
                If e.Control Then
                    Editor_Paste()
                    e.SuppressKeyPress = True
                End If

            Case Keys.Space
                If e.Control Then
                    e.SuppressKeyPress = True
                End If

        End Select

        If LV_suggestions.Visible Then
            Select Case e.KeyCode

                Case Keys.Space
                    If SelectWithSpace Then
                        e.SuppressKeyPress = True
                    End If

                Case Keys.Enter
                    If SelectWithEnter Then
                        Suggestions_AddSelected()
                        e.SuppressKeyPress = True
                    End If

                Case Keys.Up
                    ListView_ChangeSelection_Relative(LV_suggestions, -1)
                    Suggestions_ShowListViewToolTip()
                    e.SuppressKeyPress = True

                Case Keys.Down
                    ListView_ChangeSelection_Relative(LV_suggestions, 1)
                    Suggestions_ShowListViewToolTip()
                    e.SuppressKeyPress = True

                Case Keys.PageUp
                    ListView_ChangeSelection_Relative(LV_suggestions, -10)
                    Suggestions_ShowListViewToolTip()
                    e.SuppressKeyPress = True

                Case Keys.PageDown
                    ListView_ChangeSelection_Relative(LV_suggestions, 10)
                    Suggestions_ShowListViewToolTip()
                    e.SuppressKeyPress = True

            End Select

        End If
    End Sub

    Private Sub RichTextBoxCode_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles RichTextBoxCode.KeyUp
        '
        Select Case e.KeyCode
            Case Keys.Space
                If LV_suggestions.Visible Then
                    If SelectWithSpace Then
                        If Not e.Control Then
                            Suggestions_AddSelected()
                            RichTextBoxCode.SelectedText = " "
                        End If
                    End If
                Else
                    If e.Control Then
                        Suggestions_FindMembers()
                        Return
                    End If
                End If

            Case Keys.Right
                If e.Alt Then
                    Suggestions_FindMembers()
                    Return
                End If

            Case Keys.OemPeriod, Keys.Cancel, Keys.Delete, Keys.Back
                Suggestions_Editing = True
                Return

            Case Keys.Escape
                Suggestions_CloseAll()
        End Select

        If LV_suggestions.Visible Then
            Select Case e.KeyCode
                Case Keys.ControlKey
                Case Keys.Space
                Case Keys.Up, Keys.Down
                Case Else
                    LV_suggestions.Visible = False
            End Select
        End If

        ShowCaret()
    End Sub


    ' ===============================================================================================================
    '  MENU COMMANDS
    ' ===============================================================================================================
    Friend Sub Editor_NewVB()
        If Not Editor_OkToGo("New file VB") Then Exit Sub
        Editor_FileName = "NewFile.vb"
        Dim t As String = Compiler_GetEmptyNewText(Compiler_Languages.VB)
        Form_Options.AutoDetectLanguage(t, Editor_FileName)
        RichTextBoxCode.Text = t
        OldLineCount = RichTextBoxCode.Lines.Count
        Editor_NormalizeAndColorAllLines()
        Editor_UpdateAndCompile()
        '
        Undo_Clear()
        Editor_Modified = True
    End Sub
    Friend Sub Editor_NewCS()
        If Not Editor_OkToGo("New file C#") Then Exit Sub
        Editor_FileName = File_GetPath(Editor_FileName) & "NewFile.cs"
        Dim t As String = Compiler_GetEmptyNewText(Compiler_Languages.CSharp)
        Form_Options.AutoDetectLanguage(t, Editor_FileName)
        RichTextBoxCode.Text = t
        OldLineCount = RichTextBoxCode.Lines.Count
        Editor_NormalizeAndColorAllLines()
        Editor_UpdateAndCompile()
        '
        Undo_Clear()
        Editor_Modified = True
    End Sub
    Friend Sub Editor_Open(ByVal FileName As String)
        Editor_FileName = FileName
        LoadScript()
    End Sub
    Friend Sub Editor_Open()
        If Not Editor_OkToGo("File open") Then Exit Sub
        Dim OFD As System.Windows.Forms.OpenFileDialog
        OFD = New System.Windows.Forms.OpenFileDialog()
        OFD.InitialDirectory = PrepareSaveLoadFolder()
        OFD.FileName = Editor_FileName
        'OFD.RestoreDirectory = True
        OFD.DefaultExt = ".txt"
        OFD.AutoUpgradeEnabled = True
        OFD.Title = "Load script file"
        OFD.ValidateNames = False
        OFD.Multiselect = True
        OFD.Filter = "All Files|*.*|Script Files ( .vb .cs .txt )|*.vb;*.cs;*.txt"
        If OFD.ShowDialog() = DialogResult.OK Then
            If OFD.FileName.EndsWith(".exe", StringComparison.InvariantCultureIgnoreCase) Then
                System.Diagnostics.Process.Start(OFD.FileName)
            Else
                Editor_FileName = OFD.FileName
                Form_Main.m_MruList.Add(Editor_FileName)
                LoadScript()
            End If
        End If
    End Sub
    Friend Sub Editor_Save()
        SaveScript()
    End Sub
    Friend Sub Editor_SaveAs()
        Dim SFD As System.Windows.Forms.SaveFileDialog
        SFD = New System.Windows.Forms.SaveFileDialog()
        SFD.InitialDirectory = PrepareSaveLoadFolder()
        SFD.FileName = IO.Path.GetFileName(Editor_FileName)
        'SFD.RestoreDirectory = False
        SFD.DefaultExt = ".txt"
        SFD.AutoUpgradeEnabled = True
        SFD.Title = "Save script file"
        SFD.ValidateNames = False
        SFD.Filter = "Script Files ( .vb .cs .txt )|*.vb;*.cs;*.txt|All Files|*.*"
        If SFD.ShowDialog() = DialogResult.OK Then
            Editor_FileName = SFD.FileName
            SaveScript()
        End If
    End Sub

    Private Function PrepareSaveLoadFolder() As String
        PrepareSaveLoadFolder = ""
        If My.Computer.FileSystem.FileExists(Editor_FileName) Then
            PrepareSaveLoadFolder = File_GetPath(Editor_FileName)
        End If
        If Not FolderExists(PrepareSaveLoadFolder) Or _
           Not PrepareSaveLoadFolder.StartsWith(Windows.Forms.Application.StartupPath, StringComparison.InvariantCultureIgnoreCase) Then
            PrepareSaveLoadFolder = Windows.Forms.Application.StartupPath & "\Script_Examples"
        End If
    End Function

    Friend Function Editor_OkToGo(ByVal CaptionText As String) As Boolean
        If Editor_Modified Then
            Dim r As MsgBoxResult = MsgBox("Save the file before ?", MsgBoxStyle.Question Or MsgBoxStyle.YesNoCancel, CaptionText)
            If r = MsgBoxResult.Yes Then
                SaveScript()
                Return True
            End If
            If r = MsgBoxResult.No Then
                Return True
            End If
            Return False
        Else
            Return True
        End If
    End Function

    Friend Sub Editor_GotoDefinition()
        '
        Dim ww As WholeWord = FindWholeWord(RichTextBoxCode.Text, RichTextBoxCode.SelectionStart)
        If ww.Len > 0 Then
            '
            Dim startLine As Int32 = RichTextBoxCode.GetLineFromCharIndex(RichTextBoxCode.SelectionStart)
            Dim startBlock As Int32 = startLine
            Dim endBlock As Int32 = startLine
            '
            ' -------------------------------------------------------------------- find the start of this block
            Dim keywords_start As String() = {"sub", "function", "class", "module"}
            For startBlock = startLine To 0 Step -1
                Dim line As String = RichTextBoxCode.Lines(startBlock)
                Dim i1 As Int32 = STR_InstrWordarray(line, keywords_start, CompareMethod.Text) - 1
                If i1 >= 0 Then
                    If Not IsCommented(line, i1) Then Exit For
                End If
            Next
            startBlock = Math.Max(startBlock, 0)
            ' -------------------------------------------------------------------- find the end of this block
            Dim keywords_end As String() = {"end sub", "end function", "end class", "end module"}
            For endBlock = startLine To RichTextBoxCode.Lines.Count - 1
                Dim line As String = RichTextBoxCode.Lines(endBlock)
                Dim i1 As Int32 = STR_InstrWordarray(line, keywords_end, CompareMethod.Text) - 1
                If i1 >= 0 Then
                    If Not IsCommented(line, i1) Then Exit For
                End If
            Next
            endBlock = Math.Min(endBlock, RichTextBoxCode.Lines.Count - 1)
            ' -------------------------------------------------------------------- find declaration in the block
            '
            Dim declarators As String() = {"sub ", "function ", "dim ", "public ", "private ", "friend ", _
                                           "byval ", "byres ", "for ", "withevents"}
            For n As Int32 = startBlock To endBlock
                Dim line As String = RichTextBoxCode.Lines(n)

                Dim i0 As Int32 = STR_InstrWordarray(line, declarators, CompareMethod.Text) - 1
                If i0 >= 0 Then

                    Dim i1 As Int32 = InStr(i0 + 2, line, ww.Word, CompareMethod.Text) - 1
                    If i1 >= 0 Then
                        ' -------------------------------------------------------- accept only whole words
                        If FindWholeWord(line, i1).Word.Equals(ww.Word, _
                                                               StringComparison.InvariantCultureIgnoreCase) Then
                            ' ---------------------------------------------------- only if not commented
                            If Not IsCommented(line, i1) Then
                                If n <> RTB_GetFirstSelectedLine() Then
                                    RichTextBoxCode.Select(i1 + RichTextBoxCode.GetFirstCharIndexFromLine(n), ww.Len)
                                    Exit Sub
                                End If
                            End If
                        End If
                    End If
                End If
            Next
            ' -------------------------------------------------------------------- if not found then search out of the block
            '
            For n As Int32 = 0 To RichTextBoxCode.Lines.Count - 1
                '
                If n >= startBlock And n <= endBlock Then Continue For
                '
                Dim line As String = RichTextBoxCode.Lines(n)

                Dim i0 As Int32 = STR_InstrWordarray(line, declarators, CompareMethod.Text) - 1
                If i0 >= 0 Then

                    Dim i1 As Int32 = InStr(i0 + 2, line, ww.Word, CompareMethod.Text) - 1
                    If i1 >= 0 Then
                        ' -------------------------------------------------------- accept only whole words
                        If FindWholeWord(line, i1).Word.Equals(ww.Word, _
                                                               StringComparison.InvariantCultureIgnoreCase) Then
                            ' ---------------------------------------------------- only if not commented
                            If Not IsCommented(line, i1) Then
                                RichTextBoxCode.Select(i1 + RichTextBoxCode.GetFirstCharIndexFromLine(n), ww.Len)
                                Exit Sub
                            End If
                        End If
                    End If
                End If
            Next
        End If
    End Sub

    Friend Sub Editor_Cut()
        If RichTextBoxCode.SelectionLength > 0 Then
            SendTextToClipboard()
            RichTextBoxCode.SelectedText = ""
        End If
    End Sub
    Friend Sub Editor_Copy()
        If RichTextBoxCode.SelectionLength > 0 Then
            SendTextToClipboard()
        End If
    End Sub
    Private Sub SendTextToClipboard()
        Dim data As IDataObject
        data = New DataObject(DataFormats.Text, RichTextBoxCode.SelectedText.Replace(vbLf, vbCrLf))
        Clipboard.SetDataObject(data, True)
    End Sub

    Friend Sub Editor_Paste()
        If Clipboard.GetDataObject().GetDataPresent(DataFormats.Text) Then
            Editor_SetCursor_Wait()
            RTB_SuspendRedraw()
            ' ------------------------------------------------------- extract only the text from clipbbard
            RichTextBoxCode.SelectedText = Clipboard.GetText
            '
            RTB_ResumeRedraw()
            Editor_SetCursor_Normal()
        End If
    End Sub
    Friend Function Editor_CanPaste() As Boolean
        Return Clipboard.GetDataObject().GetDataPresent(DataFormats.Text)
    End Function

    Friend Sub Editor_Delete()
        RichTextBoxCode.SelectedRtf = ""
    End Sub
    Friend Sub Editor_SelectAll()
        RichTextBoxCode.SelectAll()
    End Sub


    ' ===============================================================================================================
    '  FIND - REPLACE and Options
    ' ===============================================================================================================
    Friend Sub Editor_Find()
        With Form_Replace
            .TopMost = True
            .DialogType = Form_Replace.DialogTypeValue.Find
            .OwnerCodeBox = RichTextBoxCode
            .Show()
            .txtFind.Focus()
        End With
    End Sub

    Friend Sub Editor_FindReplace()
        With Form_Replace
            .TopMost = True
            .DialogType = Form_Replace.DialogTypeValue.Replace
            .OwnerCodeBox = RichTextBoxCode
            .Show()
            .txtFind.Focus()
        End With
    End Sub

    Friend Sub Editor_ShowOptions(ByVal parent As Form)
        With Form_Options
            If .Visible Then
                .Hide()
            Else
                .Left = parent.Right - .Width
                .Top = parent.Top + 32
                .Show(parent)
            End If
        End With
    End Sub

    Friend Sub Editor_RemoveEmptyLines()
        Dim FirstSelectedLine As Int32 = RTB_GetFirstSelectedLine()
        Dim LastSelectedLine As Int32 = RTB_GetLastSelectedLine()
        If LastSelectedLine <= FirstSelectedLine Then Return
        '
        Dim s1 As String
        Dim s2 As String
        Dim s3 As String = ""
        For l As Int32 = FirstSelectedLine To LastSelectedLine
            s1 = RichTextBoxCode.Lines(l)
            s2 = s1
            s2 = Replace(s2, vbTab, "")
            s2 = Replace(s2, " ", "")
            If s2 <> "" Then
                'Debug.Print(Asc(s1).ToString)
                s3 &= s1 & vbLf
            End If
        Next
        '
        RichTextBoxCode.SelectedText = s3
        Editor_NormalizeAndColorLines(FirstSelectedLine, LastSelectedLine)
    End Sub

    Friend Sub Editor_InsertResxFile()
        '
        Dim FileName As String = GetResxFileName()
        '
        If My.Computer.FileSystem.FileExists(FileName) Then
            ' ------------------------------------------------------------------------- header
            Dim str As String
            str = "Module Resources" & vbLf & vbLf
            str &= vbTab & "Friend Function GetResource(ByVal key As String) As Object" & vbLf
            str &= vbTab & vbTab & "GetResource = Nothing" & vbLf
            str &= vbTab & vbTab & "Dim rsxr As System.Resources.ResXResourceReader" & vbLf
            str &= vbTab & vbTab & "rsxr = System.Resources.ResXResourceReader.FromFileContents(ResStr)" & vbLf
            str &= vbTab & vbTab & "Dim id As IDictionaryEnumerator = rsxr.GetEnumerator()" & vbLf
            str &= vbTab & vbTab & "For Each d As DictionaryEntry In rsxr" & vbLf
            str &= vbTab & vbTab & vbTab & "If d.key.ToString = key Then" & vbLf
            str &= vbTab & vbTab & vbTab & vbTab & "GetResource = d.value" & vbLf
            str &= vbTab & vbTab & vbTab & vbTab & "Exit For" & vbLf
            str &= vbTab & vbTab & vbTab & "End If" & vbLf
            str &= vbTab & vbTab & "Next" & vbLf
            str &= vbTab & vbTab & "rsxr.Close()" & vbLf
            str &= vbTab & "End Function" & vbLf & vbLf
            '
            str &= vbTab & "Dim ResStr As String = _" & vbLf
            ' ------------------------------------------------------------------------- resx lines
            Dim comment As Boolean = False
            Dim l As String
            Dim f As Int32 = FreeFile()
            FileOpen(f, FileName, OpenMode.Input, OpenAccess.Read, OpenShare.Shared)
            Do While Not EOF(f)
                l = LineInput(f)
                If Trim(l).StartsWith("<!--") Then comment = True

                If Not comment Then
                    If Replace(Replace(l, vbTab, ""), " ", "") <> "" Then
                        l = Replace(l, """", """""")
                        l = Trim(l)
                        str &= vbTab & """" & l & """"
                        If Not EOF(f) Then str &= " & _" & vbLf
                    End If
                End If

                If comment Then
                    If Trim(l).StartsWith("-->") Then comment = False
                End If
            Loop
            FileClose(f)
            str &= vbLf & vbLf
            str &= "End Module" & vbLf
            ' -------------------------------------------------------------------------

            Editor_SetCursor_Wait()

            RichTextBoxCode.Select(9999999, 0)
            RichTextBoxCode.SelectedText &= str

            Editor_NormalizeAndColorAllLines()

            Editor_SetCursor_Normal()
        End If
    End Sub

    Private Function GetResxFileName() As String
        Dim OFD As System.Windows.Forms.OpenFileDialog
        OFD = New System.Windows.Forms.OpenFileDialog()
        OFD.DefaultExt = ".txt"
        '
        OFD.RestoreDirectory = True
        OFD.AutoUpgradeEnabled = True
        OFD.Title = "Load script file"
        OFD.ValidateNames = False
        OFD.Filter = "Resx Files ( .resx )|*.resx|All Files|*.*"
        If OFD.ShowDialog() = DialogResult.OK Then
            Return OFD.FileName
        End If
        Return ""
    End Function


    ' ===============================================================================================================
    '   LOAD and SAVE
    ' ===============================================================================================================
    Friend Editor_FileName As String

    Friend Sub Editor_InitialReload()
        ' ----------------------------------------------------------------------------------------------------
        ' if "RichTextBoxCode.ShowSelectionMargin = True" then sometimes scrollbars are invisible at start
        '  this does a "Form Resize" and "visibilize" immediately the scroll bars 
        Form_Main.Width += 2
        ' ----------------------------------------------------------------------------------------------------
        LoadScript()
    End Sub

    Private Sub LoadScript()
        If Not My.Computer.FileSystem.FileExists(Editor_FileName) Then Exit Sub

        Editor_SetCursor_Wait()
        RTB_SuspendRedraw()
        EventsAreEnabled = False
        '
        Dim reader As System.IO.TextReader
        reader = New System.IO.StreamReader(Editor_FileName)
        Dim text As String = reader.ReadToEnd
        reader.Close()
        text = Replace(text, vbCrLf, vbLf)
        text = Replace(text, vbCr, vbLf)
        RichTextBoxCode.Text = text
        '
        OldLineCount = RichTextBoxCode.Lines.Count
        '
        Form_Options.AutoDetectLanguage(text, Editor_FileName)
        Editor_NormalizeAndColorAllLines()
        Editor_UpdateAndCompile()
        '
        RTB_ResumeRedraw()
        EventsAreEnabled = True
        Editor_SetCursor_Normal()
        '
        Undo_Clear()
    End Sub

    Private Sub SaveScript()
        If Not My.Computer.FileSystem.DirectoryExists(File_GetPath(Editor_FileName)) Then Exit Sub
        '
        Editor_SetCursor_Wait()
        '
        Dim text As String
        text = Replace(RichTextBoxCode.Text, vbLf, vbCrLf)
        Dim writer As System.IO.StreamWriter
        writer = New System.IO.StreamWriter(Editor_FileName)
        writer.Write(text)
        writer.Close()
        '
        Editor_SetCursor_Normal()
        '
        Editor_Modified = False
    End Sub


    ' ===============================================================================================================
    '  COLORED TEXT
    ' ===============================================================================================================
    Private RegexPattern_VB As String = _
        "\b(_dummy_|" & _
        "#Const|#Else|#ElseIf|#End|#End ExternalSource|#End If|#EndIf|#ExternalSource|#If|" & _
        "#Region|#End Region|Addhandler|AddressOf|Alias|And|AndAlso|Ansi|As|Assembly|Auto" & _
        "|Boolean|ByRef|Byte|ByVal|Call|Case|Case Else|Catch|CBool|CByte|CChar|CDate|CDbl" & _
        "|CDec|Char|CInt|Class|CLng|CObj|Const|CShort|CSng|CStr|CType|Date|Decimal|Declar" & _
        "e|Default|Delegate|Dim|DirectCast|Do|Do Until|Do While|Double|Each|Else|ElseIf|E" & _
        "nd|End Class|End Function|End If|End Module|End Namespace|End Property|End Selec" & _
        "t|End Sub|End While|End With|EndIf|Enum|Erase|Error|Event|Exit|Exit Function|Exi" & _
        "t Property|Exit Sub|Explicit|ExternalSource|False|Finally|For|For Each|Friend|Fu" & _
        "nction|Get|GetType|GoTo|Handles|If|Implements|Imports|In|Inherits|Integer|Interf" & _
        "ace|Is|Lib|Like|Long|Loop|Loop Until|Loop While|Me|Mod|Module|MustInherit|MustOv" & _
        "erride|MyBase|MyClass|Namespace|New|Next|Not|Nothing|NotInheritable|NotOverridab" & _
        "le|Object|Off|On|Option|Option Compare Binary|Option Compare Text|Option Explici" & _
        "t|Option Explicit Off|Option Explicit On|Option Strict|Option Strict Off|Option " & _
        "Strict On|Optional|Or|OrElse|Overloads|Overridable|Overrides|ParamArray|Preserve" & _
        "|Private|Property|Protected|Public|RaiseEvent|ReadOnly|ReDim|Redim Preserve|Regi" & _
        "on|RemoveHandler|Resume|Return|Select|Set|Shadows|Shared|Short|Single|Static|Ste" & _
        "p|Stop|Strict|String|Structure|Sub|SyncLock|Then|Throw|To|True|Try|TypeOf|Unicod" & _
        "e|Until|When|While|With|WithEvents|WriteOnly|Xor|Int32|Compare|Binary" & _
        ")\s*\b"

    Private RegexPattern_CS As String = _
        "\b(" & _
        "#define|#elif|#else|#endif|#endregion|#error|#if|#line|#region|#undef|#warning|" & _
        "abstract|as|base|bool|break|byte|case|catch|char|checked|class|const|continue|" & _
        "decimal|default|delegate|do|double|else|enum|event|explicit|extern|false|finally|" & _
        "fixed|float|for|foreach|get|goto|if|implicit|in|int|interface|internal|is|lock|" & _
        "long|namespace|new|null|object|operator|out|override|params|private|protected|" & _
        "public|readonly|ref|return|sbyte|sealed|set|short|sizeof|stackalloc|static|string|" & _
        "struct|switch|this|throw|true|try|typeof|uint|ulong|unchecked|unsafe|ushort|using|" & _
        "value|virtual|void|volatile|while" & _
        ")\s*\b"


    Private WithEvents RTF_Temp As System.Windows.Forms.RichTextBox = New System.Windows.Forms.RichTextBox()


    'Friend Sub Editor_RefreshAll()
    '    RichTextBoxCode.Rtf = RichTextBoxCode.Text
    '    Editor_NormalizeAndColorAllLines()
    '    Undo_Clear()
    'End Sub

    Friend Sub Editor_NormalizeAndColorAllLines()
        If RichTextBoxCode.Lines.Count = 0 Then Exit Sub
        '
        Dim sw1 As Diagnostics.Stopwatch = New Diagnostics.Stopwatch
        sw1.Start()
        '
        RTB_SaveSelectionAndScroll()
        '
        Undo_EnableAdd = False
        RTB_SuspendRedraw()
        '
        For l As Int32 = 0 To RichTextBoxCode.Lines.Count - 1
            RTB_NormalizeAndColorLine(l, True)
        Next
        RichTextBoxCode.BackColor = Color_Back

        RTB_RestoreSelectionAndScroll()
        '
        RTB_ResumeRedraw()
        Undo_EnableAdd = True
        '
        Form_Main.StatusLabel1.Text = "Done: " & (sw1.ElapsedMilliseconds / 1000).ToString("0.000").Replace(",", ".") & " seconds"
    End Sub

    Friend Sub Editor_NormalizeAndColorLines(ByVal firstline As Int32, ByVal lastline As Int32)
        If RichTextBoxCode.Lines.Count = 0 Then Exit Sub
        If firstline < 0 Or lastline > RichTextBoxCode.Lines.Count - 1 Then Exit Sub
        '
        RTB_SaveSelectionAndScroll()
        Undo_EnableAdd = False
        RTB_SuspendRedraw()
        '
        For l As Int32 = firstline To lastline
            RTB_NormalizeAndColorLine(l, True)
        Next
        RichTextBoxCode.BackColor = Color_Back
        '
        RTB_RestoreSelectionAndScroll()
        RTB_ResumeRedraw()
        Undo_EnableAdd = True
    End Sub

    Private Sub RTB_NormalizeAndColorLine(ByVal line As Int32, _
                                          ByVal NormalizeText As Boolean)
        '
        If line < 0 Or line >= RichTextBoxCode.Lines.Count Then Exit Sub

        ' ------------------------------------------- the background color will be overwrited so remove from marked lines
        ErrorMarker_RemoveLineFromMarkedLines(line)

        With RTF_Temp
            .Font = RichTextBoxCode.Font
            .AcceptsTab = RichTextBoxCode.AcceptsTab
            .DetectUrls = RichTextBoxCode.DetectUrls
            .WordWrap = RichTextBoxCode.WordWrap
        End With
        '
        Dim RegexPattern_Quotes As String = """+.*""+"
        Dim CommentCharacter As String = "'"
        If Compiler_Language = Compiler_Languages.CSharp Then
            RegexPattern_Quotes = """+.*""+"
            CommentCharacter = "//"
        End If

        RTB_SaveSelectionAndScroll()

        Dim lineStart As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(line)
        Dim lineLength As Int32 = RichTextBoxCode.Lines(line).Length

        ' ------------------------------------------- parse and color the code
        Dim CurrentChr As String = ""
        Dim Counter As Integer = 0
        Dim CommentCharPosition As Integer = 0
        Dim Length As Integer = 0
        Dim QuoteFlag As Boolean = False
        'Dim CommentFlag As Boolean = False
        Dim Match As Match

        With RTF_Temp
            .Clear()

            If NormalizeText Then
                .Text = STR_ConvertInitialSpacesToTabs(RichTextBoxCode.Text.Substring(lineStart, lineLength), TabSize)
            Else
                .Text = RichTextBoxCode.Text.Substring(lineStart, lineLength)
            End If

            .SelectAll()
            .SelectionColor = Color_NormalText

            ' -------------------------------------------------------------------------------- look for quoted words
            For Each Match In Regex.Matches(.Text, RegexPattern_Quotes, RegexOptions.Singleline)
                .Select(Match.Index, Match.Length)
                .SelectionColor = Color_QuotedText
                QuoteFlag = True
            Next Match

            ' -------------------------------------------------------------------------------- check for comments
            CommentCharPosition = InStr(.Text, CommentCharacter)
            If CommentCharPosition > 0 Then
                If QuoteFlag = False Then
                    ' found a comment but no quotes so don't need special processing
                    Length = Len(.Text)
                    .Select(CommentCharPosition - 1, Length)
                    .SelectionColor = Color_CommentText
                Else
                    ' we have quotes so we need to see if the comment character is contained in quotes or actually a comment
                    ' reset our flags
                    QuoteFlag = False
                    'go through one character at a time to determine if contained in quotes or comment
                    For Counter = 1 To Len(.Text)
                        CurrentChr = Mid(.Text, Counter, 1)
                        If CurrentChr = Chr(34) Then
                            ' found a quote
                            QuoteFlag = Not QuoteFlag
                        Else
                            ' allows multi character comment delimeters
                            If CurrentChr & Mid(.Text, Counter, Len(CommentCharacter) - 1) = CommentCharacter Then
                                If QuoteFlag = False Then
                                    ' found our comment and not contained in quotes
                                    .Select(Counter - 1, Len(.Text))
                                    .SelectionColor = Color_CommentText
                                End If
                            End If
                        End If
                    Next
                End If
            End If

            If CommentCharPosition = 0 Then CommentCharPosition = 9999
            If CommentCharPosition > 4 Then
                ' -------------------------------------------------------------------------------- look for keywords
                If Compiler_Language = Compiler_Languages.VB Then
                    For Each Match In Regex.Matches(.Text, _
                                                    RegexPattern_VB, _
                                                    RegexOptions.Multiline Or RegexOptions.IgnoreCase)
                        If Match.Index < CommentCharPosition Then
                            If Not InsideQuotes(Match.Index, .Text) Then
                                ' ------------------------------------------------------ replace keyword with the right case
                                Dim len As Int32 = Match.Value.Trim.Length
                                .Select(Match.Index, len)
                                Dim index As Int32 = InStr(RegexPattern_VB, "|" & Trim(Match.Value) & "|", CompareMethod.Text) + 1
                                Dim word As String = Mid(RegexPattern_VB, index, len).Replace("|", " ")
                                .SelectedText = word
                                ' ------------------------------------------------------ set keyword color
                                .Select(Match.Index, Match.Length)
                                .SelectionColor = Color_KeywordText
                            End If
                        End If
                    Next Match
                Else
                    For Each Match In Regex.Matches(.Text, RegexPattern_CS, RegexOptions.Multiline)
                        If Match.Index < CommentCharPosition Then
                            .Select(Match.Index, Match.Length)
                            .SelectionColor = Color_KeywordText
                        End If
                    Next Match
                End If
            End If


            ' ------------------------------------------------------------
            .SelectAll()
            .SelectionBackColor = Color_TextBack
        End With

        ' ---------------------------------------------------------------- copy from RTF_Temp to RichTextBoxCode
        RTB_SuspendRedraw()
        RichTextBoxCode.Select(lineStart, lineLength)
        RichTextBoxCode.SelectedRtf = RTF_Temp.SelectedRtf
        RTB_RestoreSelectionAndScroll()
        RTB_ResumeRedraw()
    End Sub

    Private Function InsideQuotes(ByVal index As Int32, ByVal text As String) As Boolean
        ' can not declare parameter "text" as ByRef (some locking subeffect?)
        InsideQuotes = False
        If index >= text.Length Then Return False
        For Counter As Int32 = 0 To index - 1
            If text(Counter) = Chr(34) Then
                InsideQuotes = Not InsideQuotes
            End If
        Next
    End Function


    ' ===============================================================================================================
    '  ListViewErrors
    ' ===============================================================================================================
    Private Sub ShowCompileErrors()
        If ListViewErrors Is Nothing Then Exit Sub
        Dim c As Color
        For Each err As CompilerError In Compiler_Results.Errors
            If err.IsWarning Then
                c = Color_WarningBack
            Else
                c = Color_ErrorBack
            End If

            Dim l As ListViewItem = New ListViewItem(Math.Max(err.Line - Compiler_HeaderLinesCount, 0).ToString())
            l.SubItems.Add(err.ErrorNumber & ": " & err.ErrorText)
            l.BackColor = c
            ListViewErrors.Items.Add(l)
        Next
        Dim s As String = ""
        If Compiler_ErrorCount > 0 Then s &= Compiler_ErrorCount.ToString & " Errors"
        If Compiler_WarningCount > 0 Then s &= "     " & Compiler_WarningCount.ToString & " Warnings"
        ListViewErrors.Columns(1).Text = s
        ListView_ResizeColumn(ListViewErrors, 0)
        ListView_ResizeColumn(ListViewErrors, 1)
        ' ------------------------------------------------------------ show the "Error List" tab
        If Compiler_ErrorCount + Compiler_WarningCount > 0 Then
            If TabReports IsNot Nothing Then
                TabReports.SelectedIndex = 0
            End If
        End If
    End Sub

    Private Sub ListViewErrors_SelectedIndexChanged(ByVal sender As Object, ByVal e As System.EventArgs)
        If ListViewErrors.SelectedItems.Count <= 0 Then Exit Sub
        Dim l As Integer = CInt(ListViewErrors.SelectedItems(0).Text) - 1
        '
        RichTextBoxCode.Select(RTB_GetLineStartIndex(l), 1)
    End Sub

    Friend Sub Editor_ShowLinkerError(ByVal s As String)
        If ListViewErrors Is Nothing Then Exit Sub
        Dim l As ListViewItem = New ListViewItem("0")
        l.SubItems.Add("Linker Error: " & s)
        l.BackColor = Color_ErrorBack
        ListViewErrors.Items.Add(l)
        ListViewErrors.Columns(1).Text = s
        ListView_ResizeColumn(ListViewErrors, 0)
        ListView_ResizeColumn(ListViewErrors, 1)
        Compiler_ErrorCount += 1
    End Sub


    ' ===============================================================================================================
    '  ErrorMarker - BACKGROUND COLOR FOR ERROR AND WARNING LINES
    ' ===============================================================================================================
    Private ErrorMarker_MarkedLines As List(Of Int32) = New List(Of Int32)

    Private Sub ErrorMarker_RemoveLineFromMarkedLines(ByVal LineNumber As Int32)
        'If ErrorMarker_MarkedLines Is Nothing Then Exit Sub
        ErrorMarker_MarkedLines.Remove(LineNumber)
    End Sub

    Private Sub ErrorMarker_AdjustMarkedLines(ByVal OldLinesCount As Int32, ByVal NewLinesCount As Int32, ByVal CaretLine As Int32)
        'If ErrorMarker_MarkedLines Is Nothing Then Exit Sub
        Dim n As Int32
        Dim DeltaLines As Int32 = NewLinesCount - OldLinesCount
        For i As Int32 = 0 To ErrorMarker_MarkedLines.Count - 1
            n = ErrorMarker_MarkedLines(i)
            If n >= CaretLine Then
                ErrorMarker_MarkedLines(i) += DeltaLines
            End If
        Next
    End Sub

    Private Sub ErrorMarker_ShowErrorLines()
        'Dim t As PrecisionTimer = New PrecisionTimer
        '
        For i As Int32 = 0 To RichTextBoxCode.Lines.Count - 1
            Dim found As Boolean = False
            For Each er As CompilerError In Compiler_Results.Errors
                If er.Line - Compiler_HeaderLinesCount = i + 1 Then ' < error lines are 1 based
                    If Not ErrorMarker_MarkedLines.Contains(i) Then
                        If er.IsWarning Then
                            ErrorMarker_MarkLine(i, Color_WarningBack)
                        Else
                            ErrorMarker_MarkLine(i, Color_ErrorBack)
                        End If
                        ErrorMarker_MarkedLines.Add(i)
                    End If
                    found = True
                    Exit For
                End If
            Next
            If Not found Then
                If ErrorMarker_MarkedLines.Contains(i) Then
                    ErrorMarker_MarkLine(i, Color_TextBack)
                    ErrorMarker_MarkedLines.Remove(i)
                End If
            End If
        Next

        'Form1.Text = t.GetTimeMicrosec.ToString
    End Sub

    Private Sub ErrorMarker_MarkLine(ByVal line As Int32, ByVal backColor As Color)
        '
        RTB_SaveSelectionAndScroll()
        RTB_SuspendRedraw()
        '
        Dim lineStart As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(line)
        Dim lineLength As Int32 = RichTextBoxCode.Lines(line).Length
        RichTextBoxCode.Select(lineStart, lineLength)
        '
        ' ------------------------------------------------ only if changed ( to facilitate undo )
        If backColor <> RichTextBoxCode.SelectionBackColor Then
            RichTextBoxCode.SelectionBackColor = backColor
        End If
        '
        RTB_RestoreSelectionAndScroll()
        RTB_ResumeRedraw()
    End Sub


    ' -----------------------------------------------------------------------------------------------------
    '  RUNTIME ERRORS
    ' -----------------------------------------------------------------------------------------------------
    Friend Sub Editor_ShowRuntimeError(ByVal strError As String)
        If TextBoxRunTimeErrors Is Nothing Then Exit Sub
        With TextBoxRunTimeErrors
            .Text &= vbCrLf
            .Text &= " ------------------------------------ Runtime Error ----------------------" & vbCrLf
            .Text &= strError & vbCrLf
            .SelectionStart = .Text.Length - 1
            .ScrollToCaret()
            .Focus()
        End With
        ' ------------------------------------------------------------ show the "Immediate" tab
        If TabReports IsNot Nothing Then
            TabReports.SelectedIndex = 1
        End If
    End Sub

    Private Sub TextBoxRunTimeErrors_TextChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles TextBoxRunTimeErrors.TextChanged, _
                                                                                                                TextBoxRunTimeErrors.Resize, _
                                                                                                                TextBoxRunTimeErrors.FontChanged
        Dim fS As Size = TextRenderer.MeasureText(TextBoxRunTimeErrors.Text, TextBoxRunTimeErrors.Font)
        Dim cS As Integer = fS.Height * fS.Width
        Dim dS As Integer = TextBoxRunTimeErrors.Size.Height * TextBoxRunTimeErrors.Size.Width
        If dS < cS Then
            TextBoxRunTimeErrors.ScrollBars = ScrollBars.Both
        Else
            TextBoxRunTimeErrors.ScrollBars = ScrollBars.None
        End If
    End Sub



    ' ===============================================================================================================
    '  UPDATE and COMPILE
    ' ===============================================================================================================
    Private OldLineCount As Int32 = -1

    Friend Sub Editor_UpdateAndCompile()
        If Editor_RunningCode Then
            Beep()
            Exit Sub
        End If
        '
        EventsAreEnabled = False
        RTB_SuspendRedraw()
        '
        ' --------------------------------------------------------------------- prepare useful parameters
        Dim LineCount As Int32 = RichTextBoxCode.Lines.Count
        Dim CaretPosition As Int32 = RichTextBoxCode.SelectionStart
        Dim CaretLine As Int32 = RichTextBoxCode.GetLineFromCharIndex(CaretPosition)

        ' --------------------------------------------------------------------- adjust marked lines
        If LineCount <> OldLineCount Then
            ErrorMarker_AdjustMarkedLines(OldLineCount, LineCount, CaretLine)
        End If


        If LineCount = OldLineCount Then
            ' ----------------------------------------------------------------- color single line
            RTB_NormalizeAndColorLine(CaretLine, False)

        ElseIf LineCount > OldLineCount Then
            ' ----------------------------------------------------------------- color many lines
            Dim firstLine As Int32 = CaretLine - (LineCount - OldLineCount)
            For i As Int32 = firstLine To CaretLine
                RTB_NormalizeAndColorLine(i, True)
                'Debug.Print(i.ToString)
            Next

        End If

        ' --------------------------------------------------------------------- compile and mark errors
        Editor_Compile()
        '
        OldLineCount = LineCount
        EventsAreEnabled = True
        RTB_ResumeRedraw()
        ShowCaret()
        RichTextBoxCode.Refresh()
        '
        ' --------------------------------------------------------------------- focus only if not ListViewSuggestions
        If Not LV_suggestions.Visible Then
            RichTextBoxCode.Focus()
        End If
    End Sub


    Friend Sub Editor_Compile()
        ' --------------------------------------------------------------------- find suggestion members
        If Suggestions_Editing Then
            Suggestions_FindMembers()
        End If
        ' --------------------------------------------------------------------- clear ListViewErrors
        If ListViewErrors IsNot Nothing Then ListViewErrors.Items.Clear()
        ' --------------------------------------------------------------------- compile
        Compiler_CompileText(RichTextBoxCode.Text, "", True, False)
        ' --------------------------------------------------------------------- editor show errors
        ShowCompileErrors()
        ' --------------------------------------------------------------------- show errors
        ErrorMarker_ShowErrorLines()
    End Sub

    Friend Sub Editor_CompileExeFile(ByVal flagRun As Boolean)
        ' --------------------------------------------------------------------- clear ListViewErrors
        If ListViewErrors IsNot Nothing Then ListViewErrors.Items.Clear()
        ' --------------------------------------------------------------------- compile
        Dim FileName As String = IO.Path.ChangeExtension(Editor_FileName, ".exe")
        Compiler_CompileText(RichTextBoxCode.Text, FileName, False, False)
        ' --------------------------------------------------------------------- editor show errors
        ShowCompileErrors()
        ' --------------------------------------------------------------------- show errors
        ErrorMarker_ShowErrorLines()
        ' --------------------------------------------------------------------- run exe file
        If flagRun Then
            Compiler_RunExeFile(FileName)
        End If
    End Sub

    Friend Sub Editor_CompileDll()
        ' --------------------------------------------------------------------- clear ListViewErrors
        If ListViewErrors IsNot Nothing Then ListViewErrors.Items.Clear()
        ' --------------------------------------------------------------------- compile
        Dim FileName As String = IO.Path.ChangeExtension(Editor_FileName, ".dll")
        Compiler_CompileText(RichTextBoxCode.Text, FileName, False, True)
        ' --------------------------------------------------------------------- editor show errors
        ShowCompileErrors()
        ' --------------------------------------------------------------------- show errors
        ErrorMarker_ShowErrorLines()
    End Sub


    ' ===============================================================================================================
    '  UNDO e REDO
    ' ===============================================================================================================
    Friend Function Editor_CanUndo() As Boolean
        Return Undo_Index >= 2
    End Function

    Friend Function Editor_CanRedo() As Boolean
        Return Undo_Index <= UndoBuffer.Count - 1
    End Function

    Friend Sub Editor_Undo()
        If Undo_Index < 2 Then Exit Sub
        '
        RTB_SuspendRedraw()
        EventsAreEnabled = False
        Undo_EnableAdd = False
        '
        Undo_Index -= 1
        Dim undo As UndoEntry

        undo = UndoBuffer(Undo_Index - 1)
        RichTextBoxCode.Rtf = undo.RTF_Text

        'undo = UndoBuffer(Undo_Index)

        RichTextBoxCode.Select(undo.FirstVisibleChar, 0)
        RichTextBoxCode.ScrollToCaret()

        RichTextBoxCode.Select(undo.SelectionStart, undo.SelectionLength)
        ErrorMarker_MarkedLines = undo.MarkedLines
        OldLineCount = undo.LineCount
        '
        ' --------------------------------------------------------------------- compile and mark errors
        Editor_Compile()
        ShowCaret()

        EventsAreEnabled = True
        Undo_EnableAdd = True
        RTB_ResumeRedraw()

        If Undo_Index < 2 Then
            Editor_Modified = False
        End If
    End Sub

    Friend Sub Editor_Redo()
        If Undo_Index > UndoBuffer.Count - 1 Then Exit Sub
        '
        Editor_Modified = True
        '
        RTB_SuspendRedraw()
        EventsAreEnabled = False
        Undo_EnableAdd = False
        '
        Dim undo As UndoEntry = UndoBuffer(Undo_Index)
        RichTextBoxCode.Rtf = undo.RTF_Text

        RichTextBoxCode.Select(undo.FirstVisibleChar, 0)
        RichTextBoxCode.ScrollToCaret()

        RichTextBoxCode.Select(undo.SelectionStart, undo.SelectionLength)
        ErrorMarker_MarkedLines = undo.MarkedLines
        OldLineCount = undo.LineCount

        Undo_Index += 1
        ' --------------------------------------------------------------------- compile and mark errors
        Editor_Compile()
        ShowCaret()
        '
        EventsAreEnabled = True
        Undo_EnableAdd = True
        RTB_ResumeRedraw()
        '
        Editor_Modified = True
    End Sub


    ' -------------------------------------------------------------- private undo helpers
    Private Structure UndoEntry
        Dim ActionName As String
        Dim RTF_Text As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
        Dim MarkedLines As List(Of Int32)
        Dim LineCount As Int32
    End Structure
    Private UndoBuffer As List(Of UndoEntry) = New List(Of UndoEntry)
    Private Undo_EnableAdd As Boolean = True
    Private Undo_Index As Int32

    Private Sub Undo_Clear()
        UndoBuffer.Clear()
        Undo_Index = 0
        Undo_AddEntry()
        Editor_Modified = False
    End Sub
    Private Sub Undo_AddEntry()
        If Not Undo_EnableAdd Then Exit Sub
        '
        Dim newentry As UndoEntry
        newentry.FirstVisibleChar = RichTextBoxCode.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = RichTextBoxCode.SelectionStart
        newentry.SelectionLength = RichTextBoxCode.SelectionLength
        newentry.MarkedLines = ErrorMarker_MarkedLines
        newentry.LineCount = OldLineCount
        newentry.ActionName = ""
        newentry.RTF_Text = RichTextBoxCode.Rtf

        While UndoBuffer.Count > Undo_Index
            UndoBuffer.RemoveAt(UndoBuffer.Count - 1)
        End While

        UndoBuffer.Add(newentry)
        Undo_Index += 1
        'Debug.Print(UndoBuffer.Count.ToString)

        Editor_Modified = True
    End Sub

    ' ===============================================================================================================
    '   "SUSPEND REDRAW" HELPER
    ' ===============================================================================================================
    Private Const WM_SETREDRAW As Int32 = 11
    Public Sub SuspendRedraw(ByVal cnt As Control)
        Dim msgSuspendUpdate As Message
        msgSuspendUpdate = Message.Create(cnt.Handle, WM_SETREDRAW, IntPtr.Zero, IntPtr.Zero)
        NativeWindow.FromHandle(cnt.Handle).DefWndProc(msgSuspendUpdate)
    End Sub
    Public Sub ResumeRedraw(ByVal cnt As Control)
        Dim msgResumeUpdate As Message
        msgResumeUpdate = Message.Create(cnt.Handle, WM_SETREDRAW, New IntPtr(1), IntPtr.Zero)
        NativeWindow.FromHandle(cnt.Handle).DefWndProc(msgResumeUpdate)
        cnt.Invalidate()
    End Sub

    ' ===============================================================================================================
    '   RichTextBoxCode - enhanced "SuspendRedraw" that can be used "nested"
    ' ===============================================================================================================
    Private RTB_redraw_suspended As Int32 = 0
    Private Sub RTB_SuspendRedraw()
        If RTB_redraw_suspended <= 0 Then
            SuspendRedraw(RichTextBoxCode)
            RTB_redraw_suspended = 1
        Else
            RTB_redraw_suspended += 1
        End If
    End Sub
    Private Sub RTB_ResumeRedraw()
        RTB_redraw_suspended -= 1
        If RTB_redraw_suspended = 0 Then
            ResumeRedraw(RichTextBoxCode)
            RichTextBoxCode.Refresh()
            'RichTextBoxCode.Invalidate()
        End If
    End Sub


    ' ===============================================================================================================
    '   RichTextBoxCode - SAVE and RESTORE selection and scroll     ( scroll aligned with the first text line )
    ' ===============================================================================================================
    Private RTB_OldSelectionStart As Int32
    Private RTB_OldSelectionLength As Int32
    Private RTB_OldFirstChar As Int32
    Private RTB_selection_saved As Int32 = 0
    Private Sub RTB_SaveSelectionAndScroll()
        If RTB_selection_saved <= 0 Then
            RTB_OldSelectionStart = RichTextBoxCode.SelectionStart
            RTB_OldSelectionLength = RichTextBoxCode.SelectionLength
            RTB_OldFirstChar = RichTextBoxCode.GetCharIndexFromPosition(New Point(1, 1))
            'Debug.Print(OLD_FirstChar.ToString)
            RTB_selection_saved = 1
        Else
            RTB_selection_saved += 1
        End If
    End Sub
    Private Sub RTB_RestoreSelectionAndScroll()
        RTB_selection_saved -= 1
        If RTB_selection_saved = 0 Then
            RichTextBoxCode.Select(RTB_OldFirstChar, 0)
            RichTextBoxCode.ScrollToCaret()
            RichTextBoxCode.Select(RTB_OldSelectionStart, RTB_OldSelectionLength)
        End If
    End Sub


    ' ===============================================================================================================
    '   RichTextBoxCode - SAVE and RESTORE - SCROLL POSITION       ( scroll is pixel aligned )
    ' ===============================================================================================================
    'Private Declare Auto Function SendMessagePoint Lib "user32" Alias "SendMessage" (ByVal hwndLock As IntPtr, _
    '                                                                             ByVal wMsg As Int32, _
    '                                                                             ByVal wParam As Int32, _
    '                                                                             ByRef pt As Point) As Int32
    'Friend Function GetScrollPosition(ByVal rtb As RichTextBox) As Point
    '    Const EM_GETSCROLLPOS As Int32 = &H400 + 221
    '    SendMessagePoint(rtb.Handle, EM_GETSCROLLPOS, 0, GetScrollPosition)
    'End Function
    'Friend Sub SetScrollPosition(ByVal rtb As RichTextBox, ByRef pt As Point)
    '    Const EM_SETSCROLLPOS As Int32 = &H400 + 222
    '    SendMessagePoint(rtb.Handle, EM_SETSCROLLPOS, 0, pt)
    'End Sub


    'Private RTB_OldScrollPosition As Point
    'Private Sub RTB_SaveScrollPosition()
    '    RTB_OldScrollPosition = GetScrollPosition(RichTextBoxCode)
    'End Sub
    'Private Sub RTB_RestoreScrollPosition()
    '    SetScrollPosition(RichTextBoxCode, RTB_OldScrollPosition)
    'End Sub



    ' ===============================================================================================================
    '  RICH TEXT BOX HELPER FUNCTIONS
    ' ===============================================================================================================
    'Private Function RTB_GetLineLength(ByVal line As Int32) As Int32
    '    If line < 0 Or line > RichTextBoxCode.Lines.Count - 1 Then
    '        Return 0
    '    Else
    '        Return RichTextBoxCode.Lines(line).Length
    '    End If
    'End Function

    Private Function RTB_GetLineStartIndex(ByVal line As Int32) As Int32
        If line < 0 Or line > RichTextBoxCode.Lines.Count - 1 Then
            Return 0
        Else
            Return RichTextBoxCode.GetFirstCharIndexFromLine(line)
        End If
    End Function

    Private Sub RTB_SetTabPositions()
        Dim ar(31) As Int32
        For i As Int32 = 0 To 31
            ar(i) = CInt(i * TabSize * RichTextBoxCode.Font.Height / 2)
        Next
        RichTextBoxCode.SelectionTabs = ar
    End Sub

    Private Function RTB_GetWordCount() As Int32
        Dim Counter As Int32
        Dim Match As System.Text.RegularExpressions.Match
        For Each Match In System.Text.RegularExpressions.Regex.Matches(RichTextBoxCode.Text, "\w+")
            Counter += 1
        Next
        Return Counter
    End Function

    Private Function RTB_GetFirstSelectedLine() As Int32
        Return RichTextBoxCode.GetLineFromCharIndex(RichTextBoxCode.SelectionStart)
    End Function
    Private Function RTB_GetLastSelectedLine() As Int32
        Return RichTextBoxCode.GetLineFromCharIndex(RichTextBoxCode.SelectionStart + RichTextBoxCode.SelectionLength)
    End Function

    Private Sub RTB_SelectLines(ByVal FirstLine As Int32, ByVal LastLine As Int32)
        Dim FirstChar As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(FirstLine)
        Dim LastChar As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(LastLine) + RichTextBoxCode.Lines(LastLine).Length - 1
        RichTextBoxCode.Select(FirstChar, LastChar - FirstChar)
    End Sub

    Private Function RTB_FindIdentifierStart(ByVal index As Int32) As Int32
        Do
            If index < 1 Then Exit Do
            index -= 1
            Dim ch As Char = RichTextBoxCode.Text(index)
            If Not Char.IsLetterOrDigit(ch) And ch <> "_" And ch <> "." Then
                index += 1
                Exit Do
            End If
        Loop
        Return index
    End Function

    Private Function RTB_FindIdentifierEnd(ByVal index As Int32) As Int32
        Do
            If index < 0 Or index > RichTextBoxCode.Text.Count - 1 Then Return -1
            Dim ch As Char = RichTextBoxCode.Text(index)
            If Not Char.IsLetterOrDigit(ch) And ch <> "_" And ch <> "." Then
                index -= 1
                Exit Do
            End If
            index += 1
        Loop
        Return index
    End Function

    Private Sub RTB_IndentSelectedLines(ByVal increment As Int32)
        If EventsAreEnabled Then
            EventsAreEnabled = False
            RTB_SuspendRedraw()
            '
            Undo_AddEntry()
            '
            Dim FirstSelectedLine As Int32 = RTB_GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = RTB_GetLastSelectedLine()
            '
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RichTextBoxCode.Lines(i)
                Dim ntabs As Int32 = STR_CountInitialTabs(lineText, TabSize) + increment
                If ntabs >= 0 Then
                    Dim start As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(i)
                    RichTextBoxCode.Select(start, STR_CountInitialChars(lineText))
                    RichTextBoxCode.SelectedText = StrDup(ntabs, vbTab)
                End If
            Next
            '
            RTB_SelectLines(FirstSelectedLine, LastSelectedLine)
            '
            RTB_ResumeRedraw()
            EventsAreEnabled = True
        End If
    End Sub

    Private Sub RTB_CommentSelectedLines(ByVal comment As Boolean)
        If EventsAreEnabled Then
            EventsAreEnabled = False
            RTB_SuspendRedraw()
            '
            Undo_AddEntry()
            '
            Dim CommentColumn As Int32 = -1
            '
            Dim FirstSelectedLine As Int32 = RTB_GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = RTB_GetLastSelectedLine()
            '
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RichTextBoxCode.Lines(i)
                Dim firstch As Int32 = STR_CountInitialChars(lineText)
                If firstch >= 0 Then
                    If CommentColumn < 0 Then CommentColumn = firstch
                    Dim start As Int32 = RichTextBoxCode.GetFirstCharIndexFromLine(i)

                    If comment Then
                        RichTextBoxCode.Select(start + CommentColumn, 0)
                        RichTextBoxCode.SelectedText = Editor_CommentChars
                    Else
                        RichTextBoxCode.Select(start + firstch, Editor_CommentChars.Length)
                        If RichTextBoxCode.SelectedText = Editor_CommentChars Then
                            RichTextBoxCode.SelectedText = ""
                        End If
                    End If
                End If
            Next

            Editor_NormalizeAndColorLines(FirstSelectedLine, LastSelectedLine)
            '
            RTB_SelectLines(FirstSelectedLine, LastSelectedLine)
            '
            RTB_ResumeRedraw()
            '
            ' --------------------------------------------------------------------- compile and mark errors
            Editor_Compile()
            '
            EventsAreEnabled = True
        End If
    End Sub



    ' ===============================================================================================================
    '  STRING HELPER FUNCTIONS
    ' ===============================================================================================================
    Private Function STR_CountInitialChars(ByVal s As String) As Int32
        Dim CharCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : CharCount += 1
                Case " " : CharCount += 1
                Case Else : Exit For
            End Select
        Next
        Return CharCount
    End Function

    Private Function STR_CountInitialTabs(ByVal s As String, ByVal TabSize As Int32) As Int32
        Dim SpaceCount As Int32 = 0
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case " " : SpaceCount += 1
                Case Else : Exit For
            End Select
        Next
        Return TabCount + SpaceCount \ TabSize
    End Function

    Private Function STR_LtrimTabsAndSpaces(ByVal s As String) As String
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab
                Case " "
                Case Else
                    Return s.Substring(i)
            End Select
        Next
        Return ""
    End Function

    Private Function STR_ConvertInitialSpacesToTabs(ByVal s As String, ByVal TabSize As Int32) As String
        Dim ntabs As Int32 = STR_CountInitialTabs(s, TabSize)
        Dim s2 As String = STR_LtrimTabsAndSpaces(s)
        If s2 <> "" Then
            Return StrDup(ntabs, vbTab) & s2
        Else
            Return s
        End If
    End Function

    Public Function File_GetPath(ByVal str As String) As String
        Try
            Return IO.Path.GetDirectoryName(str) & "\"
        Catch
            Return str
        End Try
    End Function

    Public Function FolderExists(ByVal FolderName As String) As Boolean
        If FolderName Is Nothing Then Return False
        If FolderName.Length < 2 Then Return False
        FolderName = LCase(FolderName)
        Select Case FolderName
            Case "a:\", "b:\", "c:\", "d:\", "e:\", "f:\", "g:\", "h:\", "i:\", "j:\", "k:\"
                Return True
        End Select
        Return My.Computer.FileSystem.DirectoryExists(FolderName)
    End Function

    ' zero if not found - 1 if at beginning
    Friend Function STR_InstrWordarray(ByVal str As String, ByVal ar() As String, ByVal compare As CompareMethod) As Int32
        Dim i As Int32
        For Each s As String In ar
            i = InStr(str, s, compare)
            If i > 0 Then
                Return i
            End If
        Next
        Return 0
    End Function


    ' ===============================================================================================================
    '  LISTVIEW HELPER FUNCTIONS
    ' ===============================================================================================================
    Friend Sub ListView_ChangeSelection_Relative(ByVal lv As ListView, ByVal delta As Int32)
        If lv.SelectedIndices.Count = 0 Then Exit Sub
        Dim i As Int32 = lv.SelectedIndices(0) + delta
        If i >= lv.Items.Count Then i = lv.Items.Count - 1
        If i < 0 Then i = 0
        lv.Items(i).Selected = True
        lv.Items(i).EnsureVisible()
    End Sub
    Private Sub ListView_ResizeColumn(ByRef lview As ListView, ByVal c As Int32)
        lview.Columns(c).AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent)
        Dim cw As Int32 = lview.Columns(c).Width
        lview.Columns(c).AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize)
        Dim hw As Int32 = lview.Columns(c).Width
        If (cw > hw) Then
            lview.Columns(c).AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent)
        Else
            lview.Columns(c).AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize)
        End If
    End Sub
    Private Function ListView_FindItemIndex_WithText(ByVal lv As ListView, ByVal text As String) As Int32
        For Each lvi As ListViewItem In lv.Items
            If lvi.Text.StartsWith(text, StringComparison.InvariantCultureIgnoreCase) Then
                Return lvi.Index
            End If
        Next
        Return -1
    End Function


    ' ===============================================================================================================
    '   SUGGESTIONS
    ' ===============================================================================================================
    Private Suggestions_Editing As Boolean = False
    Private Suggestions_StartIndex As Int32
    Private Suggestions_EndIndex As Int32
    Private Suggestions_Wcomplete As String
    Private Suggestions_Wstart As String
    Private Suggestions_Wend As String
    Private Suggestions_Wpointed As String
    Private Suggestions_Words() As String

    Private Sub Suggestions_FindMembers()
        Suggestions_StartIndex = RTB_FindIdentifierStart(RichTextBoxCode.SelectionStart)
        Suggestions_EndIndex = RTB_FindIdentifierEnd(RichTextBoxCode.SelectionStart)
        If Suggestions_EndIndex < Suggestions_StartIndex Then Return

        ' ------------------------------------------------------------------------ return if the char is after a comment
        If IsCommented(RichTextBoxCode.SelectionStart) Then Return

        Suggestions_Wcomplete = RichTextBoxCode.Text.Substring(Suggestions_StartIndex, Suggestions_EndIndex - Suggestions_StartIndex + 1)
        Suggestions_Words = Split(Suggestions_Wcomplete, ".")
        Suggestions_Wstart = Suggestions_Words(0)

        If Suggestions_Wstart.Count > 0 Then

            Suggestions_Wend = Suggestions_Words(Suggestions_Words.Count - 1)

            ' --------------------------------------------------------------------- recompile without errors
            If Compiler_ErrorCount > 0 Then
                Compiler_CompileText(RichTextBoxCode.Text, "", True, False, True)
            End If

            Dim strType As String = FindType_in_CompiledAssembly(Suggestions_Wstart)

            If strType <> "" Then

                Dim t As Type = FindType_in_CurrentDomainAssemblies(strType)

                If t IsNot Nothing Then
                    SuspendRedraw(LV_suggestions)
                    Suggestions_FillListViewWithTypeMembers(t)
                    Suggestions_Show()
                    ResumeRedraw(LV_suggestions)
                    Exit Sub
                End If

            End If
        End If
        ' ---------------------------------------------------------------------- if not found then close all
        Suggestions_CloseAll()
    End Sub


    Private Sub Suggestions_FindInfo()

        If Suggestions_Editing Then Exit Sub

        Dim mousepoint As Point = RichTextBoxCode.PointToClient(Cursor.Position)
        Dim IndicatedCharIndex As Int32 = RichTextBoxCode.GetCharIndexFromPosition(mousepoint)

        Suggestions_StartIndex = RTB_FindIdentifierStart(IndicatedCharIndex)
        Suggestions_EndIndex = RTB_FindIdentifierEnd(IndicatedCharIndex)
        'RichTextBoxCode.Select(Suggestions_StartIndex, Suggestions_EndIndex - Suggestions_StartIndex + 1)

        ' -------------------------------------------------------------------- return if after a comment
        If IsCommented(IndicatedCharIndex) Then Return


        If mousepoint.X > RichTextBoxCode.GetPositionFromCharIndex(Suggestions_EndIndex).X + 8 * Zoom Then
            Suggestions_CloseAll()
            Exit Sub
        End If

        Suggestions_Wcomplete = RichTextBoxCode.Text.Substring(Suggestions_StartIndex, Suggestions_EndIndex - Suggestions_StartIndex + 1)
        Suggestions_Words = Split(Suggestions_Wcomplete, ".")
        Suggestions_Wpointed = FindWholeWord(RichTextBoxCode.Text, IndicatedCharIndex).Word
        Suggestions_Wstart = Suggestions_Words(0)

        If Suggestions_Wstart.Count > 0 Then

            Suggestions_Wend = Suggestions_Words(Suggestions_Words.Count - 1)

            ' --------------------------------------------------------------------- recompile without errors
            If Compiler_ErrorCount > 0 Then
                Compiler_CompileText(RichTextBoxCode.Text, "", True, False, True)
            End If

            Dim strType As String = FindType_in_CompiledAssembly(Suggestions_Wstart)

            If strType <> "" Then

                Dim t As Type = FindType_in_CurrentDomainAssemblies(strType)
                If t IsNot Nothing Then
                    strType = t.ToString
                End If

                Dim p As Point = RichTextBoxCode.GetPositionFromCharIndex(Suggestions_StartIndex)

                Suggestions_ShowInfoToolTip(p, strType)

                Exit Sub

            End If
        End If
        ' ---------------------------------------------------------------------- if not found then close all
        Suggestions_CloseAll()
    End Sub


    Private Function FindType_in_CompiledAssembly(ByVal SearchWord As String) As String

        If Compiler_Results Is Nothing Then Return ""
        If Compiler_Results.Errors.HasErrors Then Return ""

        ' ----------------------------------------------------------------
        Dim asm As Assembly = Compiler_Results.CompiledAssembly
        'Dim AsmName As String = asm.ManifestModule.ScopeName

        'Debug.Print(AsmName)
        'Dim tt As Type = asm.GetType("Script", False, True)  ' <<<< FASTER !!!

        ' ---------------------------------------------------------------- asm 
        For Each t As Type In asm.GetTypes()

            'Debug.WriteLine(t.ToString)

            'If t.ToString = "Script" Then
            If Not t.ToString.StartsWith("My.") Then

                ' -------------------------------------------------------- binding flags 
                Dim b As BindingFlags
                'b = BindingFlags.NonPublic Or _
                '    BindingFlags.Public Or _
                '    BindingFlags.Static Or _
                '    BindingFlags.IgnoreCase Or _
                '    BindingFlags.Instance

                'b = b Or BindingFlags.ExactBinding

                b = BindingFlags.NonPublic Or _
                    BindingFlags.Public Or _
                    BindingFlags.Static Or _
                    BindingFlags.Instance


                '' -------------------------------------------------------- variables,    etc...
                'For Each f As FieldInfo In t.GetFields(b)
                '    If f.Name = SearchWord Then
                '        'Debug.Print(" ----------- FIELD: " & f.ToString)
                '        'Return f.ToString.Split(New Char() {" "c})(0)
                '        Return f.FieldType.FullName
                '    End If
                'Next

                ' -------------------------------------------------------- sub, functions, forms etc.. 
                For Each mi As MemberInfo In t.GetMembers(b)

                    'Debug.Print(" ----------- MEMBER: " & mi.ToString)

                    If mi.Name.Equals(SearchWord, StringComparison.InvariantCultureIgnoreCase) Then
                        Return mi.ToString.Split(New Char() {" "c})(0)
                    End If

                    'If InStr(mi.ToString, SearchWord, CompareMethod.Text) > 0 Then
                    '    Return mi.ToString.Split(New Char() {" "c})(0)
                    'End If


                    If mi.MemberType = MemberTypes.Method Then

                        ' --------------------------------------------------------------- METHOD PARAMETERS
                        '
                        'http://www.java2s.com/Code/CSharp/Development-Class/GetMethodParamemters.htm
                        Dim mb As MethodBase = TryCast(mi, MethodBase)
                        For Each pi As ParameterInfo In mb.GetParameters
                            'Debug.Print(" ----------- PARAMETERS: " & pi.ToString)
                            If pi.Name.Equals(SearchWord, StringComparison.InvariantCultureIgnoreCase) Then
                                Return pi.ToString.Split(New Char() {" "c})(0)
                            End If
                        Next


                        ' -------------------------------------------------------------- METHOD LOCAL VARS
                        '
                        'http://msdn.microsoft.com/en-us/library/system.reflection.localvariableinfo.aspx
                        'http://msdn.microsoft.com/en-us/library/system.reflection.localvariableinfo.aspx
                        'Dim _methodinfo As MethodInfo = t.GetMethod(mi.Name)
                        'If _methodinfo IsNot Nothing Then
                        '    Dim mbody As MethodBody = _methodinfo.GetMethodBody()
                        '    If mbody IsNot Nothing Then
                        '        For Each localvarinfo As LocalVariableInfo In mbody.LocalVariables
                        '            'Debug.Print(" ----------- LOCALVARS: " & localvarinfo.ToString())
                        '            'If localvarinfo.LocalType.FullName Then
                        '            'Return localvarinfo.LocalType.FullName
                        '        Next
                        '    End If
                        'End If


                    End If


                Next

                ' -------------------------------------------------------- 
                'For Each t2 As System.Type In t.GetNestedTypes()

                '    Debug.Print(" ----------- NESTED: " & t2.ToString)

                '    For Each mi2 As MemberInfo In t2.GetMembers(b)
                '        If mi2.Name.Equals(SearchWord, StringComparison.InvariantCultureIgnoreCase) Then

                '            Return mi2.ToString.Split(New Char() {" "c})(0)
                '        End If
                '    Next
                'Next


                ' -------------------------------------------------------- variables etc.. 
                'For Each pi As PropertyInfo In t.GetProperties(b)
                '    If pi.Name = SearchWord Then
                '        'Debug.Print(" ----------- MEMBER: " & mi.ToString)
                '        Return pi.ToString.Split(New Char() {" "c})(0)
                '        'Return pi.GetType.FullName
                '    End If
                'Next


            End If
        Next

        'Return "System.Windows.Forms.Control.Controls"
        'Return "System.Windows.Forms.Control"
        'Return "System.Windows.Forms.Controls"
        'Return "System.Windows.Forms.Form.Controls"

        'Return "System.Windows.Forms.FormStartPosition"
        'Return "Windows.Forms.FormStartPosition"
        'Return "Forms.FormStartPosition"

        Dim w As String
        If Suggestions_Words.Count > 2 Then
            w = Suggestions_Words(Suggestions_Words.Count - 2)
        Else
            w = Suggestions_Wstart
        End If
        If w <> "System" Then
            w = "." & w
        End If

        Return w
    End Function

    Private Function FindType_in_CurrentDomainAssemblies(ByVal type_name_to_search As String) As Type

        If type_name_to_search.EndsWith("&") Then
            ' ---------------------------------------------------------------- this identifier is "By Reference"
            type_name_to_search = Replace(type_name_to_search, "&", "")
        End If

        For Each ASM As Assembly In AppDomain.CurrentDomain.GetAssemblies()
            If ASM.Location <> "" Then

                ' ------------------------------------------------------------ assemblies
                's &= ASM.FullName & vbCrLf

                ' ------------------------------------------------------------ modules
                'For Each m As [Module] In ASM.GetModules(True)
                '    s &= "  " & m.Name & vbCrLf
                'Next

                'If ASM.FullName.StartsWith("System.Windows.Forms") Then
                '    Beep()
                'End If

                ' ------------------------------------------------------------ types
                ' types are 2400 / exportedTypes only 1200
                'Debug.Print("-------------------------" & ASM.ToString)
                'Debug.Print(ASM.GetExportedTypes().Count.ToString)
                'Debug.Print(ASM.GetTypes().Count.ToString)

                For Each t As Type In ASM.GetExportedTypes()
                    If t.ToString.EndsWith(type_name_to_search, StringComparison.InvariantCultureIgnoreCase) Then
                        Return t
                    End If
                Next

            End If
        Next

        ' ---------------------------------------------------------------- if not found retray with the last word
        'For Each ASM As Assembly In AppDomain.CurrentDomain.GetAssemblies()
        '    If ASM.Location <> "" Then

        '        For Each t As Type In ASM.GetExportedTypes()

        '            For Each mi As MemberInfo In t.GetMembers()
        '                If mi.Name = Suggestions_WWend.Word Then
        '                    Suggestions_FillListViewWithTypeMembers(t)
        '                    Suggestions_Show()
        '                    Exit Sub
        '                End If
        '            Next
        '        Next
        '    End If
        'Next

        Return Nothing

    End Function


    Private Sub Suggestions_FillListViewWithTypeMembers(ByVal t As Type)

        LV_suggestions.Items.Clear()

        ' ------------------- member types -------------------------------------------------------------
        'Constructor    Specifies that the member is a constructor, representing a ConstructorInfo member. Hexadecimal value of 0x01. 
        'Event          Specifies that the member is an event, representing an EventInfo member. Hexadecimal value of 0x02. 
        'Field          Specifies that the member is a field, representing a FieldInfo member. Hexadecimal value of 0x04. 
        'Method         Specifies that the member is a method, representing a MethodInfo member. Hexadecimal value of 0x08. 
        'Property       Specifies that the member is a property, representing a PropertyInfo member. Hexadecimal value of 0x10. 
        'TypeInfo       Specifies that the member is a type, representing a TypeInfo member. Hexadecimal value of 0x20. 
        'Custom         Specifies that the member is a custom member type. Hexadecimal value of 0x40. 
        'NestedType     Specifies that the member is a nested type, extending MemberInfo. 
        'All            Specifies all member types. 

        'Dim b As BindingFlags = BindingFlags.NonPublic Or _
        '                        BindingFlags.Public Or _
        '                        BindingFlags.Static Or _
        '                        BindingFlags.IgnoreCase Or _
        '                        BindingFlags.Instance

        For Each mi As MemberInfo In t.GetMembers()
            If mi.Name.StartsWith("get_") Or _
               mi.Name.StartsWith("set_") Or _
               mi.Name.StartsWith("add_") Or _
               mi.Name.StartsWith("remove_") Or _
               mi.Name.StartsWith("op_") Or _
               mi.Name.StartsWith(".ctor") Then Continue For

            If Suggestions_Wend <> Suggestions_Wstart And _
               Not mi.Name.StartsWith(Suggestions_Wend, StringComparison.InvariantCultureIgnoreCase) Then Continue For

            If Not LV_suggestions.Items.ContainsKey(mi.Name) Then
                '
                Dim strPrefix As String = ""
                Dim intIcon As Int32
                '
                Select Case mi.MemberType.ToString
                    Case "Constructor"
                        intIcon = 0
                        strPrefix = "Constructor"
                    Case "Event"
                        intIcon = 1
                        strPrefix = "Event" & vbCrLf
                        Dim ei As EventInfo = t.GetEvent(mi.Name)
                        If ei IsNot Nothing Then
                            strPrefix &= " - " & ei.ToString & vbCrLf
                        End If
                    Case "Field"
                        intIcon = 2
                        strPrefix = "Field"
                        'Dim fi As FieldInfo = t.GetField(mi.Name)
                        'If fi IsNot Nothing Then
                        '    If fi.IsPublic Then strPrefix &= "Public "
                        '    If fi.IsStatic Then strPrefix &= "Static "
                        '    If fi.IsInitOnly Then strPrefix &= "Readonly "
                        'End If
                    Case "Method"
                        intIcon = 3
                        strPrefix = "Method" & vbCrLf
                        For Each mti As MethodInfo In t.GetMethods
                            If mti.Name.Equals(mi.Name, StringComparison.InvariantCultureIgnoreCase) Then
                                strPrefix &= " - " & mti.ToString & vbCrLf
                            End If
                        Next
                    Case "Property"
                        intIcon = 4
                        strPrefix = "Property" & vbCrLf
                        Dim pi As PropertyInfo = t.GetProperty(mi.Name)
                        If pi IsNot Nothing Then
                            If Not pi.CanWrite Then
                                strPrefix = "Readonly " & strPrefix
                                intIcon = 5
                            End If
                            strPrefix &= " - " & pi.ToString & vbCrLf
                        End If
                    Case "TypeInfo"
                        strPrefix = "TypeInfo"
                        intIcon = 6
                    Case "Custom"
                        strPrefix = "Custom"
                        intIcon = 7
                    Case "NestedType"
                        strPrefix = "NestedType"
                        intIcon = 8
                    Case Else
                        strPrefix = "UNKNOWN MEMBERTYPE"
                        intIcon = 9
                End Select
                '
                Dim lvi As ListViewItem = LV_suggestions.Items.Add(mi.Name, mi.Name, intIcon)
                lvi.ToolTipText = strPrefix & " ( " & t.ToString & "." & mi.Name & " )"
            End If
        Next
    End Sub


    ' ===============================================================================================================
    '  LIST VIEW SUGGESTIONS - SHOW
    ' ===============================================================================================================
    Private Sub Suggestions_Show()
        '
        If LV_suggestions.Items.Count = 0 Then
            Suggestions_CloseAll()
            Exit Sub
        End If

        Dim ListEight As Int32
        With LV_suggestions
            If .Items.Count < 10 Then
                ListEight = .Items.Count * 17 + 2
            Else
                ListEight = 174
            End If
        End With

        Dim p As Point
        With RichTextBoxCode
            p = .GetPositionFromCharIndex(.SelectionStart)
            p = .PointToScreen(p)
            p.X -= 24
            p.Y += 24
            If p.Y > Screen.PrimaryScreen.Bounds.Height - (ListEight + 50) Then
                p.Y -= (ListEight + 36)
            End If
            p = Editor_Form.PointToClient(p)
        End With

        With LV_suggestions
            .Left = p.X
            .Top = p.Y
            .Height = ListEight

            .AutoResizeColumn(0, ColumnHeaderAutoResizeStyle.ColumnContent)
            .Width = .Columns(0).Width + 24

            'Debug.Print(endWord)
            Dim ItemToSelect As Int32 = ListView_FindItemIndex_WithText(LV_suggestions, Suggestions_Wend)
            If ItemToSelect < 0 Then
                ItemToSelect = .Items.Count \ 2
            End If

            .EnsureVisible(Math.Max(ItemToSelect - 4, 0))
            .EnsureVisible(Math.Min(ItemToSelect + 5, .Items.Count - 1))
            .EnsureVisible(ItemToSelect)
            .Items(ItemToSelect).Selected = True
            .Items(ItemToSelect).Focused = True

            .BringToFront()
            .Show()

            Suggestions_Editing = True
        End With

        Suggestions_ShowListViewToolTip()
    End Sub


    Private ToolTip1 As ToolTip = New ToolTip

    Private Sub LV_suggestions_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles LV_suggestions.Click
        Suggestions_ShowListViewToolTip()
    End Sub
    Private Sub LV_suggestions_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles LV_suggestions.MouseClick
        'beep()
    End Sub
    Private Sub LV_suggestions_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles LV_suggestions.MouseDoubleClick
        Suggestions_AddSelected()
    End Sub
    Private Sub LV_suggestions_GotFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles LV_suggestions.GotFocus
        Application.DoEvents()
        RichTextBoxCode.Focus()
    End Sub

    Private Sub Suggestions_AddSelected()
        If LV_suggestions.SelectedItems.Count = 1 Then
            RTB_SelectCompleteWord()
            RichTextBoxCode.SelectedText = LV_suggestions.SelectedItems(0).Text
            Suggestions_CloseAll()
        End If
    End Sub

    Private Sub Suggestions_ShowListViewToolTip()
        If LV_suggestions.SelectedItems.Count = 1 Then
            Dim item As ListViewItem = LV_suggestions.SelectedItems(0)
            'ToolTip1.IsBalloon = True
            ToolTip1.UseAnimation = True
            ToolTip1.UseFading = True
            ToolTip1.BackColor = Color.FromArgb(255, 250, 235)
            ToolTip1.ReshowDelay = 0
            ToolTip1.Show(item.ToolTipText, _
                          CType(LV_suggestions, IWin32Window), _
                          item.Position.X + LV_suggestions.Width, _
                          item.Position.Y)
        End If
    End Sub
    Private Sub Suggestions_ShowInfoToolTip(ByVal p As Point, ByVal text As String)
        ToolTip1.UseAnimation = True
        ToolTip1.UseFading = True
        ToolTip1.BackColor = Color.FromArgb(255, 250, 235)
        ToolTip1.ReshowDelay = 0
        ' --------------------------- show two times to overwrite immediately the old text
        ToolTip1.Show(text, _
                      CType(RichTextBoxCode, IWin32Window), _
                      p.X, _
                      p.Y + 20)
        ToolTip1.Show(text, _
                      CType(RichTextBoxCode, IWin32Window), _
                      p.X, _
                      p.Y + 20)
    End Sub
    Private Sub LV_suggestions_HideToolTip()
        ToolTip1.Hide(CType(LV_suggestions, IWin32Window))
    End Sub

    Private Sub Suggestions_CloseAll()
        LV_suggestions.Hide()
        LV_suggestions_HideToolTip()
        LV_suggestions.Items.Clear()
        Suggestions_Editing = False
        RichTextBoxCode.Focus()
        Editor_Compile()
    End Sub

End Module





