' ================================================================================================
'  Theremino System - ReleDriver
' ================================================================================================
Const NumRele as Int32 = 16					' Rele number			- Numero di RELE (8,16,24...)
Const LabelSize as int32 = 32				' Label SIZE			- Larghezza delle caselle

Const SlotEnable as Int32 = 1				' Slot ENABLE			- Slot del filo ENABLE
Const SlotData as Int32 = 2					' Slot DATA				- Slot del filo DATA
Const SlotClock as Int32 = 3				' Slot CLOCK			- Slot del filo CLOCK

Const DelayMillisec as Int32 = 1			' Time for each bit		- Millisecondi per ogni bit

Dim ColorUsed as Color = Color.Orange		' Color for RELE USED	- Colore per RELE USATO
Dim ColorON as Color = Color.Red			' Color for RELE ON		- Colore per RELE ACCESO
Dim ColorArmed as Color = Color.Yellow		' Color for ARMED		- Colore per CASELLA ARMATA
Dim ColorUnused as Color = Color.White		' Color for UNUSED		- Colore per CASELLA NON USATA
Dim ColorBack as Color = Color.LightYellow	' Color for BACK		- Colore per SFONDO FINESTRA 
Dim ColorText as Color = Color.Black		' Color for TEXT		- Colore per TESTO



' ================================================================================================
'  - ENG - THE SIMPLIFIED MANAGEMENT ENDS HERE - MODIFY THE FOLLOWING CAREFULLY
' ------------------------------------------------------------------------------------------------
'  - ITA - LA GESTIONE SEMPLIFICATA FINISCE QUI - MODIFICARE LA PARTE SEGUENTE CON ATTENZIONE
' ================================================================================================

' ------------------------------------------------------------------------------------------------
'  Main loop
' ------------------------------------------------------------------------------------------------
Sub Main(ByVal s() As String)
	if Form1 IsNot Nothing then return
	Form1 = New form
	InitializeComponents
	MemoryMappedFile_Init
	While Form1.Visible
		Application.DoEvents
		Threading.Thread.Sleep(10)
	End While
End Sub

dim ValidKeyDown as boolean = True
Sub Form1_KeyDown(sender As Object, e As System.Windows.Forms.KeyEventArgs) Handles Form1.KeyDown
	if e.KeyCode = Keys.Space and ValidKeyDown then
		ValidKeyDown = false
		Pressed_Start
	End If
End Sub
Sub Form1_KeyUp(sender As Object, e As System.Windows.Forms.KeyEventArgs) Handles Form1.KeyUp
	if e.KeyCode = Keys.Space then
		ValidKeyDown = true
	End If
End Sub

Sub Button1_GotFocus(sender As Object, e As System.EventArgs) Handles Button1.GotFocus
	Label1.Focus
End Sub

Sub Button1_MouseDown(sender As Object, e As MouseEventArgs) Handles Button1.MouseDown
	Pressed_Start
End Sub

Sub PauseMS(Byval mS As Int32)
	Threading.Thread.Sleep(mS)
End Sub

' ---------------------------------------------------------------------- User Interface Components
Dim WithEvents Form1 As Form
Dim Label1 As Label = new Label
Dim LabelArray(NumRele - 1) as Label
Dim WithEvents Button1 As Button = new Button

Sub InitializeComponents()
	Application.EnableVisualStyles
	form1.SuspendLayout
	' ------------------------------------------------------------------ Labels
	InitializeLabels
	' ------------------------------------------------------------------ Form
	Form1.Text = "Theremino Script - Rele Driver V2"
	Form1.ClientSize = New System.Drawing.Size(LabelArray(NumRele - 1).right + 12, 230)
	Form1.StartPosition = FormStartPosition.CenterScreen
	Form1.FormBorderStyle = FormBorderStyle.Fixed3D
	Form1.MaximizeBox = False
	Form1.BackColor = ColorBack
	Form1.KeyPreview = True
	' ------------------------------------------------------------------ Label 1
	Label1.BorderStyle = BorderStyle.Fixed3D
	Label1.ForeColor = ColorText
	Label1.BackColor = ColorUnused 
	Label1.Location = New Point(10, 10)
	Label1.Font = New Font("Arial", 12 )
	Label1.Size = New Size(Form1.ClientSize.Width - 20, 21)
	Form1.Controls.Add(Label1)
	' ------------------------------------------------------------------ Button 1
	Button1.BackColor = ColorUnused
	Button1.ForeColor = ColorText
	Button1.Location = New Point(10, 172)
	Button1.Font = New Font("Arial", 12)
	Button1.Size = New Size(Form1.ClientSize.Width - 20, 50)
	Button1.Text = "START BUTTON" & vbcrlf & "(or press SPACE on KEYBOARD)"
	Button1.TabStop = false
	Form1.Controls.Add(Button1)
	' ------------------------------------------------------------------ 
	Form1.ResumeLayout
	Form1.PerformLayout()
	Form1.Show
End Sub

Public Delegate Sub ClickHandlerDelegate (sender As Object, e As EventArgs)
Dim instance As New ClickHandlerDelegate(AddressOf ClickHandler)
Sub ClickHandler(sender As Object, e As EventArgs)
	Dim l as Label = CType(sender, Label)
	dim n as int32 = CInt(Val(l.Text))
	if l.BackColor <> ColorON and l.BackColor <> ColorUsed Then
		Beep
		if l.BackColor <> ColorArmed Then
			l.BackColor = ColorArmed 
			Label1.Text = "Rele " & n.ToString & " ready"
		Else
			l.BackColor = ColorUnused 
			Label1.Text = "Rele " & n.ToString & " disabled"
		End If
	End If
End Sub

Sub InitializeLabels()
	For i as int32 = 0 to LabelArray.Length - 1
		LabelArray(i) = New Label
		Form1.Controls.Add(LabelArray(i))
		With LabelArray(i)
			.BackColor = ColorUnused 
			.BorderStyle = BorderStyle.Fixed3D
			.Size = New Size(LabelSize, 120)
			.Location = New Point(10 + (LabelSize + 2) * i, 40)
			.Font = New Font("Arial", 12)
			.Text = (i + 1).tostring
			.TextAlign = System.Drawing.ContentAlignment.TopCenter
			AddHandler .Click, AddressOf ClickHandler
		End With
	Next
End Sub

Sub Pressed_Start
	Beep
	dim Sequence as string = ""
	For i as Int32 = 0 to NumRele -1
		With LabelArray(i)
			If .BackColor = ColorON Then
				.BackColor = ColorUsed
			End If
			If .BackColor = ColorArmed Then
				Sequence = Sequence & "1"
				.BackColor = ColorON
			Else
				Sequence = Sequence & "0"
			End If 
		End With
	Next
	SendShiftSequence(Sequence)
End Sub

Sub SendShiftSequence (Byval Sequence as String)
	Label1.text = "Shift sequence = " & Sequence
	Form1.Refresh
	WriteSlot(SlotEnable, 0)
	For Each c as Char in Sequence
		If c = "0" then
			WriteSlot(SlotData, 0)
		Else
			WriteSlot(SlotData, 1000)
		End If
		WriteSlot(SlotClock, 1000)
		PauseMS(DelayMillisec)
		WriteSlot(SlotClock, 0)
		PauseMS(DelayMillisec)
	Next
	WriteSlot(SlotEnable, 1000)
	WriteSlot(SlotData, 0)
End Sub

