' ================================================================================================
'  Theremino System - ReleDriver V5
' ------------------------------------------------------------------------------------------------
'  - ENG - 
'  This applicazione must be used with "Theremino Switcher V2"
'  Data are sent to a shift register that controls a bank of relays (or other devices)
'  STORE, DATA, CLOCK And 5Volt signals are sent via a Master Module and four shielded wires
'  to receiver modules at great distance (hundred of meters)
'  A suitable hardware here: http://www.theremino.com/hardware/outputs/actuators#releswitcherv2
' ------------------------------------------------------------------------------------------------
'  - ITA - 
'  Questa applicazione deve essere usata con il "Theremino Switcher V2"
'  I dati vengono inviati a uno shift register che pilota un banco di relè (o altro) 
'  I segnali STORE, DATA, CLOCK e 5VOLT sono inviati via Master e cavo schermato a quattro fili
'  a moduli ricevitori situati a grande distanza (centinaia di metri).
'  Un esempio di hardware qui: http://www.theremino.com/hardware/outputs/actuators#releswitcherv2
' ================================================================================================

Dim NumRele As Int32 = 32 					' Rele number			- Numero di RELE (8,16,24...)
Dim LabelsPerRow As Int32 = 16				' Labels for each row	- Caselle per riga(8,16,24...)
Dim LabelSizeX As Int32 = 40				' Label SIZE X			- Larghezza delle caselle
Dim LabelSizeY As Int32 = 40				' Label SIZE Y			- Altezza delle caselle

Dim SlotStore As Int32 = 1					' Slot ENABLE			- Slot del filo STORE
Dim SlotData As Int32 = 2					' Slot DATA				- Slot del filo DATA
Dim SlotClock As Int32 = 3					' Slot CLOCK			- Slot del filo CLOCK

Dim DelayMillisec As Int32 = 4				' Time For Each bit		- Millisecondi per ogni bit

Dim WatchDogEnabled As Boolean = False		' Watchdog enabled      - Abilitazione del WatchDog
Dim WatchDogPulsesMillisec As Int32 = 100	' Watchdog pulse len    - Durata impulsi WatchDog

Dim ColorUsed As Color = Color.Orange		' Color For RELE USED	- Colore per RELE' USATO
Dim ColorON As Color = Color.Red			' Color For RELE On		- Colore per RELE' ACCESO
Dim ColorArmed As Color = Color.Yellow		' Color For ARMED		- Colore per CASELLA ARMATA
Dim ColorUnused As Color = Color.White		' Color For UNUSED		- Colore per CASELLA NON USATA
Dim ColorBack As Color = Color.LightYellow	' Color For BACK		- Colore per SFONDO FINESTRA 
Dim ColorText As Color = Color.Black		' Color For TEXT		- Colore per TESTO



' ================================================================================================
'  - ENG - THE SIMPLIFIED MANAGEMENT ENDS HERE - MODIFY THE FOLLOWING CAREFULLY
' ------------------------------------------------------------------------------------------------
'  - ITA - LA GESTIONE SEMPLIFICATA FINISCE QUI - MODIFICARE LA PARTE SEGUENTE CON ATTENZIONE
' ================================================================================================

' ------------------------------------------------------------------------------------------------
'  Main Loop
' ------------------------------------------------------------------------------------------------
Sub Main(ByVal s() As String)
	If Form1 IsNot Nothing Then Return
	Form1 = New form
	InitializeComponents
	MemoryMappedFile_Init
	timeBeginPeriod(1)
	SendEmptyDataString
	StartThread()
	While Form1.Visible
		Application.DoEvents
		Threading.Thread.Sleep(10)
	End While
End Sub

Sub Form1_Closing(sender As Object, e As CancelEventArgs) Handles Form1.Closing
	SendEmptyDataString
	timeEndPeriod(1)
	StopThread()
End Sub

'Dim ValidKeyDown As Boolean = True
'Sub Form1_KeyDown(sender As Object, e As System.Windows.Forms.KeyEventArgs) Handles Form1.KeyDown
'	If e.KeyCode = Keys.Space And ValidKeyDown Then
'		ValidKeyDown = False
'		If Command <> "" Then Return
'		Command = "START"
'	End If
'End Sub
'Sub Form1_KeyUp(sender As Object, e As System.Windows.Forms.KeyEventArgs) Handles Form1.KeyUp
'	If e.KeyCode = Keys.Space Then
'		ValidKeyDown = True
'	End If
'End Sub

Sub Button1_GotFocus(sender As Object, e As System.EventArgs) Handles Button1.GotFocus
	Label1.Focus
End Sub

Sub Button1_MouseDown(sender As Object, e As MouseEventArgs) Handles Button1.MouseDown
	If Command <> "" Then Return
	Command = "START"
End Sub

Sub Button2_MouseDown(sender As Object, e As MouseEventArgs) Handles Button2.MouseDown
	If Command <> "" Then Return
	Command = "RESET"
End Sub

Sub Button3_MouseDown(sender As Object, e As MouseEventArgs) Handles Button3.MouseDown
	If Command <> "" Then Return
	Command = "SEQUENCE"
End Sub

Sub TrackBar1_MouseDown(sender As Object, e As EventArgs) Handles TrackBar1.ValueChanged
	If Command <> "" Then Return
	Label1.Text = "Progress initial time = " & SpeedToString(TrackBar1.value)
End Sub

Sub TrackBar2_MouseDown(sender As Object, e As EventArgs) Handles TrackBar2.ValueChanged
	If Command <> "" Then Return
	Label1.Text = "Progress final time = " & SpeedToString(TrackBar2.value)
End Sub

Function SpeedToString(ByVal speed As Int32) As String
	Dim sec As Single = SpeedToDelay(speed) 
	sec += NumRele * DelayMillisec * 2
	sec /= 1000f
	If sec >= 2 Then
		Return sec.ToString("0", System.Globalization.CultureInfo.InvariantCulture) & " Sec"
	Else
		Return sec.ToString("0.0", System.Globalization.CultureInfo.InvariantCulture) & " Sec"
	End If
End Function


' ---------------------------------------------------------------------- User Interface Components
Dim WithEvents Form1 As Form
Dim Label1 As Label = New Label
Dim LabelArray(- 1) As Label
Dim WithEvents Button1 As Button = New Button
Dim WithEvents Button2 As Button = New Button
Dim WithEvents Button3 As Button = New Button
Dim WithEvents TrackBar1 As TrackBar = New TrackBar
Dim WithEvents TrackBar2 As TrackBar = New TrackBar
Dim Label2 As Label = New Label
Dim Label3 As Label = New Label

Sub InitializeComponents()
	Application.EnableVisualStyles
	form1.SuspendLayout
	' ------------------------------------------------------------------ Params correction
	NumRele = 8 * (NumRele \ 8) 
	If NumRele < 8 Then NumRele = 8
	LabelsPerRow = 8 * (LabelsPerRow \ 8)
	If LabelsPerRow < 8 Then LabelsPerRow = 8
	' ------------------------------------------------------------------ Labels
	ReDim LabelArray(NumRele - 1)
	InitializeLabels
	Dim x As Int32 = LabelArray(NumRele - 1).right
	Dim y As Int32 = LabelArray(NumRele - 1).bottom
	' ------------------------------------------------------------------ Form
	Form1.Text = "Theremino Script - Rele Driver V5"
	Form1.ClientSize = New System.Drawing.Size(x + 12, y + 120)
	Form1.StartPosition = FormStartPosition.CenterScreen
	Form1.FormBorderStyle = FormBorderStyle.Fixed3D
	Form1.MaximizeBox = False
	Form1.BackColor = ColorBack
	Form1.KeyPreview = True
	' ------------------------------------------------------------------ Label 1
	Label1.BorderStyle = BorderStyle.Fixed3D
	Label1.ForeColor = ColorText
	Label1.BackColor = ColorUnused 
	Label1.Location = New Point(10, 10)
	Label1.Font = New Font("Arial", 12)
	Label1.Size = New Size(x - 8, 21)
	Form1.Controls.Add(Label1)
	' ------------------------------------------------------------------ Button 1
	Button1.BackColor = ColorUnused
	Button1.ForeColor = ColorText
	Button1.Location = New Point(10, y + 8)
	Button1.Font = New Font("Arial", 12)
	Button1.Size = New Size(x - 98, 50)
	Button1.Text = "Start" ' "START" & vbcrlf & "(or key SPACE)"
	Button1.TabStop = False
	Form1.Controls.Add(Button1)
	' ------------------------------------------------------------------ Button 2
	Button2.BackColor = ColorUnused
	Button2.ForeColor = ColorText
	Button2.Location = New Point(x - 82, y + 8)
	Button2.Font = New Font("Arial", 12)
	Button2.Size = New Size(86, 50)
	Button2.Text = "Reset"
	Button2.TabStop = False
	Form1.Controls.Add(Button2)
	' ------------------------------------------------------------------ Button 3
	Button3.BackColor = ColorUnused
	Button3.ForeColor = ColorText
	Button3.Location = New Point(160, y + 64)
	Button3.Font = New Font("Arial", 12)
	Button3.Size = New Size(x - 308, 50)
	Button3.Text = "Progressive sequence"
	Button3.TabStop = False
	Form1.Controls.Add(Button3)
	' ------------------------------------------------------------------ TrackBar 1
	TrackBar1.Location = New Point(10, y + 64)
	TrackBar1.AutoSize = False
	TrackBar1.Size = New Size(150, 30)
	TrackBar1.Orientation = Orientation.Horizontal
	TrackBar1.TickStyle = TickStyle.TopLeft
	TrackBar1.Maximum = 9
	TrackBar1.Minimum = 1
	TrackBar1.TickFrequency = 1
	TrackBar1.LargeChange = 1
	TrackBar1.Value = 5
	Form1.Controls.Add(TrackBar1)
	' ------------------------------------------------------------------ TrackBar 2
	TrackBar2.Location = New Point(x - 140, y + 64)
	TrackBar2.AutoSize = False
	TrackBar2.Size = New Size(150, 30)
	TrackBar2.Orientation = Orientation.Horizontal
	TrackBar2.TickStyle = TickStyle.TopLeft
	TrackBar2.Maximum = 9
	TrackBar2.Minimum = 1
	TrackBar2.TickFrequency = 1
	TrackBar2.LargeChange = 1
	TrackBar2.Value = 5
	Form1.Controls.Add(TrackBar2)
	' ------------------------------------------------------------------ Label 2
	Label2.BorderStyle = BorderStyle.None
	Label2.ForeColor = ColorText
	Label2.BackColor = color.Transparent 
	Label2.Location = New Point(16, y + 94)
	Label2.Font = New Font("Arial", 11)
	Label2.Size = New Size(160, 21)
	Label2.Text = "Slow                 Fast"
	Form1.Controls.Add(Label2)
	' ------------------------------------------------------------------ Label 3
	Label3.BorderStyle = BorderStyle.None
	Label3.ForeColor = ColorText
	Label3.BackColor = color.Transparent 
	Label3.Location = New Point(x - 132, y + 94)
	Label3.Font = New Font("Arial", 11)
	Label3.Size = New Size(160, 21)
	Label3.Text = "Slow                 Fast"
	Form1.Controls.Add(Label3)
	' ------------------------------------------------------------------ Form Show
	Form1.ResumeLayout
	Form1.PerformLayout()
	Form1.Show
End Sub

Sub InitializeLabels()
	For i As Int32 = 0 To LabelArray.Length - 1
		LabelArray(i) = New Label
		Form1.Controls.Add(LabelArray(i))
		With LabelArray(i)
			.ForeColor = ColorText
			.BackColor = ColorUnused 
			.BorderStyle = BorderStyle.Fixed3D
			.Size = New Size(LabelSizeX, LabelSizeY)
			.Location = New Point(10 + (LabelSizeX + 2) * (i Mod LabelsPerRow), _
								40 + (LabelSizeY + 2) * (i \ LabelsPerRow))
			.Font = New Font("Arial", 12)
			.Text = (i + 1).tostring
			.TextAlign = System.Drawing.ContentAlignment.TopCenter
			Addhandler .Click, AddressOf ClickHandler
		End With
	Next
End Sub

Sub ClickHandler(sender As Object, e As EventArgs)
	If Command <> "" Then Return
	Dim l As Label = CType(sender, Label)
	Dim n As Int32 = CInt(Val(l.Text))
	If l.BackColor <> ColorON And l.BackColor <> ColorUsed Then
		If l.BackColor <> ColorArmed Then
			l.BackColor = ColorArmed 
			Label1.Text = "Rele " & n.ToString & " ready"
		Else
			l.BackColor = ColorUnused 
			Label1.Text = "Rele " & n.ToString & " disabled"
		End If
	End If
End Sub

'Function MouseButtonLeftPressed() As Boolean
'	Return (Control.MouseButtons And Windows.Forms.MouseButtons.Left) <> Windows.Forms.MouseButtons.None
'End Function


' ===============================================================
'  WINDOWS SCHEDULER PRECISION
' =============================================================== 
Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32


' ===============================================================
'  WORKER THREAD
' =============================================================== 
Dim WorkerThread As Threading.Thread
Sub StartThread()
	' --------------------------------------------------------- asynchronously start the Thread 
	WorkerThread = New Threading.Thread(AddressOf ExecuteThread)
	' --------------------------------------------------------- make the thread as background thread.
	WorkerThread.IsBackground = True
	' --------------------------------------------------------- set the Priority of the thread.
	WorkerThread.Priority = Threading.ThreadPriority.highest
	' --------------------------------------------------------- start the thread.
	WorkerThread.Start()
End Sub
Sub StopThread()
	WorkerThread.Abort
End Sub
Sub ExecuteThread
	Do
		Select Case Command
			Case "START"
				ExecCommand_Start()
				Command = ""
			Case "RESET"
				ExecCommand_Reset()
				Command = ""
			Case "SEQUENCE"
				ExecCommand_Sequence()
				Command = ""
			Case Else
				WaitWithWatchDog(2 * WatchDogPulsesMillisec)
		End Select
		Threading.Thread.Sleep(1)
	Loop Until False
End Sub

Sub WaitWithWatchDog(ByVal ms As Int32)
	If WatchDogEnabled Then
		While ms >= 2 * WatchDogPulsesMillisec
			WriteSlot(SlotStore, 1000)
			Threading.Thread.Sleep(WatchDogPulsesMillisec)
			WriteSlot(SlotStore, 0)
			Threading.Thread.Sleep(WatchDogPulsesMillisec)
			ms -= 2 * WatchDogPulsesMillisec
		End While
	End If
	Threading.Thread.Sleep(ms)
End Sub


' ===============================================================
'  COMMANDS EXECUTION
' =============================================================== 
Dim Command As String = ""

Sub DisableControls
	Button1.Enabled = False
	Button2.Enabled = False
	Button3.Enabled = False
	TrackBar1.Enabled = False
	TrackBar2.Enabled = False
	'For i As Int32 = 0 To LabelArray.Length - 1
	'	LabelArray(i).Enabled = False
	'Next
End Sub

Sub EnableControls
	Button1.Enabled = True
	Button2.Enabled = True
	Button3.Enabled = True
	TrackBar1.Enabled = True
	TrackBar2.Enabled = True
	'For i As Int32 = 0 To LabelArray.Length - 1
	'	LabelArray(i).Enabled = True
	'Next
End Sub

Sub ExecCommand_Reset
	DisableControls
	SendEmptyDataString
	EnableControls
	form1.invalidate
End Sub

Sub ExecCommand_Start
	DisableControls
	Dim DataString As String = ""
	For i As Int32 = 0 To NumRele -1
		With LabelArray(i)
			If .BackColor = ColorON Then
				.BackColor = ColorUsed
			End If
			If .BackColor = ColorArmed Then
				DataString = DataString & "1"
				.BackColor = ColorON
			Else
				DataString = DataString & "0"
			End If 
		End With
	Next
	SendDataString(DataString)
	DisableLastRele
	EnableControls
	form1.invalidate
End Sub

Sub ExecCommand_Sequence
	DisableControls
	Dim NumSteps As Int32 = 0
	For i As Int32 = 0 To NumRele -1
		With LabelArray(i)
			If .BackColor = ColorON Then
				.BackColor = ColorUsed
			End If
			If .BackColor = ColorArmed Then
				NumSteps += 1
			End If
		End With
	Next
	Dim Speed As Single = trackbar1.value
	Dim DeltaSpeed As Single
	If NumSteps > 2 Then
		DeltaSpeed= (trackbar2.value - trackbar1.value) / (NumSteps - 2f)
	End If
	For s As Int32 = 1 To NumSteps
		Dim DataString As String = ""
		For i As Int32 = 0 To NumRele -1
			With LabelArray(i)
				If .BackColor = ColorON Then
					.BackColor = ColorUsed
				End If
				If .BackColor = ColorArmed Then
					DataString = DataString & "1"
					.BackColor = ColorON
					Exit For
				Else
					DataString = DataString & "0"
				End If
			End With
		Next
		DataString = DataString & StrDup(NumRele - DataString.Length, "0")
		SendDataString(DataString)
		If s < NumSteps Then
			WaitWithWatchDog(SpeedToDelay(Speed))
			Speed += DeltaSpeed
		End If
	Next
	DisableLastRele
	EnableControls
	form1.Invalidate
End Sub

Function SpeedToDelay(ByVal speed As Single) As Int32
	Select Case speed
		Case Is < 1.5: Return 20000
		Case Is < 2.0: Return 15000
		Case Is < 2.5: Return 10000
		Case Is < 3.0: Return 8000
		Case Is < 3.5: Return 5000
		Case Is < 4.0: Return 3000
		Case Is < 4.5: Return 2000
		Case Is < 5.0: Return 1500
		Case Is < 5.5: Return 1200
		Case Is < 6.0: Return 1000
		Case Is < 6.5: Return 800
		Case Is < 7.0: Return 600
		Case Is < 7.5: Return 400
		Case Is < 6.0: Return 200
		Case Is < 8.5: Return 100
		Case Else: Return 0
	End Select
End Function

Sub DisableLastRele
	WaitWithWatchDog(1000)
	Dim DataString As String = ""
	For i As Int32 = 0 To NumRele -1
		With LabelArray(i)
			If .BackColor = ColorON Then
				.BackColor = ColorUsed
			End If
		End With
		DataString = DataString & "0"
	Next
	SendDataString(DataString)
End Sub

Sub SendEmptyDataString
	Dim DataString As String = ""
	For i As Int32 = 0 To NumRele -1
		With LabelArray(i)
			.BackColor = ColorUnused
		End With
		DataString = DataString & "0"
	Next
	SendDataString(DataString) 
End Sub

Sub SendDataString(ByVal DataString As String)
	Label1.text = "Data bits: " & DataString
	Form1.Refresh
	WriteSlot(SlotStore, 0)
	For i As Int32 = DataString.length - 1 To 0 Step -1
		If DataString(i) = "0" Then
			WriteSlot(SlotData, 0)
		Else
			WriteSlot(SlotData, 1000)
		End If
		WriteSlot(SlotClock, 0)
		Threading.Thread.Sleep(DelayMillisec)
		WriteSlot(SlotClock, 1000)
		Threading.Thread.Sleep(DelayMillisec)
	Next
	WriteSlot(SlotStore, 1000)
	Threading.Thread.Sleep(DelayMillisec)
	WriteSlot(SlotData, 0)
	WriteSlot(SlotClock, 0)
	WriteSlot(SlotStore, 0)
End Sub


