' ================================================================================================
' THEREMINO Geiger Simulator
' ================================================================================================
' ------------------------------------------------------------------------------------------------
' - ENG -  This script simulates geiger pulses up to 10000 CPS
' ------------------------------------------------------------------------------------------------
' ------------------------------------------------------------------------------------------------
' - ITA - Questo script simula gli impulsi generati dai geiger fino a 10 000 CPS
' ------------------------------------------------------------------------------------------------


' ------------------------------------------------------------------------------------------------
'  - ENG - This function is executed only at the program start - Do here all the initializatons
' ------------------------------------------------------------------------------------------------
'  - ITA - Questa funzione viene esguita solo alla partenza - Fare qui tutte le inizializzazioni
' ------------------------------------------------------------------------------------------------
Sub Initialize
	Timer1.Interval = 100
	TrackBar1.Value = 2
	giGeigerPulseMillisec = 1000
	gfGeigerValue = ReadSlot(1)
End Sub


' - ENGLISH --------------------------------------------------------------------------------------
'  By default this function is called 20 times per second
'  It is possible to change the calling time with: "Timer1.Interval = xxx" (xxx is time in ms)
'  All the main window controls updates must be implemented here
'  All the frequent slot updates must be done here
'  No PAUSES here, only fast executions.
' ------------------------------------------------------------------------------------------------

' - ITALIAN --------------------------------------------------------------------------------------
'  Per default funzione viene chiamata 20 volte al secondo
'  E' possibile modificare il tempo di chiamate con: "Timer1.Interval = xxx" (xxx = tempo in ms)
'  Tutte le visualizzazioni dei dati nei controlli della finestra principale devono avvenire qui
'  Se si devono leggere o scrivere i valori degli slot con grande frequenza e' bene farlo qui
'  Niente PAUSE qui dentro, solo esecuzioni piu' rapide possibile
' ------------------------------------------------------------------------------------------------
Sub FastExecution
	Label2.Text = "Pulse time = " & giGeigerPulseMillisec.ToString("0.#") & " mS"
	Label3.Text = "Counter = " & gfGeigerValue.ToString("0")
End Sub


' - ENGLISH --------------------------------------------------------------------------------------
'  Here you can perform any kind of slow operation. The user interface runs anyway.
' ------------------------------------------------------------------------------------------------
' - ITALIAN --------------------------------------------------------------------------------------
'  Qui si possono fare operazioni lente e loop infiniti, senza bloccare l'intera applicazione.
' ------------------------------------------------------------------------------------------------
Dim gfGeigerValue As Double
Dim giGeigerPulseMillisec As Double
' Max integer stored in a double without rounding errors: 2^53 or 9x10^15 (300000 years in mSec)
Dim oldmillisec As Double 
Dim newmillisec As Double
Dim sw As StopWatch = New StopWatch
Sub PausableExecution()
	Label1.Text = "Executing geiger simulation - Output SLOT is 1"
	Label4.Text = "Change pulse time with the Track Bar"
	sw.start
	Do
		' ------------------------------------------------------------- calculate pulses for the elapsed time
		newmillisec = sw.Elapsed.TotalMilliseconds
		gfGeigerValue +=  (newmillisec - oldmillisec) / giGeigerPulseMillisec 
		oldmillisec = newmillisec
		' ------------------------------------------------------------- limit pulses to 16 bit range
		If gfGeigerValue >= 65536 Then gfGeigerValue -= 65536
		WriteSlot(1, CInt(gfGeigerValue))
		PauseMS(5)
	Loop
End Sub


' ------------------------------------------------------------------------------------------------
'  - ENG - Functions are called by the user controls events
' ------------------------------------------------------------------------------------------------
'  - ITA - Funzioni chiamate dagli eventi causati dai controlli utente
' ------------------------------------------------------------------------------------------------
Dim Times() As Single = {5000, 2000, 1000, 500, 200, 100, 50, 20, 10, 5, 2, 1, 0.5, 0.2, 0.1}
Sub TrackBar1_Scroll(value As Single)	
	giGeigerPulseMillisec = Times(CInt(value))
End Sub


' ================================================================================================
'  - ENG - THE SIMPLIFIED MANAGEMENT ENDS HER - MODIFY THE FOLLOWING CAREFULLY
' ------------------------------------------------------------------------------------------------
'  - ITA - LA GESTIONE SEMPLIFICATA FINISCE QUI - MODIFICARE LA PARTE SEGUENTE CON ATTENZIONE
' ================================================================================================


Sub Main(ByVal s() As String)
	If Form1 IsNot Nothing Then Return
	Form1 = New Form
	InitializeComponents
	MemoryMappedFile_Init
	timeBeginPeriod(1)
	StartThread
	Initialize
	While Form1.Visible
		Application.DoEvents
		Threading.Thread.Sleep(10)
	End While
End Sub

Sub Form1_Closing(sender As Object, e As CancelEventArgs) Handles Form1.Closing
	timeEndPeriod(1)
	Timer1.Stop
	objThread.Abort
End Sub


Private Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
Private Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32

Sub StartThread()
	' --------------------------------------------------------- asynchronously start the Thread 
	objThread = New Threading.Thread(AddressOf ExecuteThread)
	' --------------------------------------------------------- make the thread as background thread.
	objThread.IsBackground = True
	' --------------------------------------------------------- set the Priority of the thread.
	objThread.Priority = Threading.ThreadPriority.AboveNormal
	' --------------------------------------------------------- start the thread.
	objThread.Start()
End Sub

Sub ExecuteThread
	Do
		PausableExecution
		PauseMS(1)
	Loop Until False
End Sub

Sub PauseMS(ByVal mS As Int32)
	Threading.Thread.Sleep(mS)
End Sub

Sub Timer1_Tick(ByVal sender As Object, ByVal e As EventArgs) Handles Timer1.Tick
	FastExecution
End Sub

Sub TrackBar1_Scroll(ByVal sender As Object, ByVal e As EventArgs) Handles TrackBar1.Scroll
	TrackBar1_Scroll(TrackBar1.Value)
End Sub

' ---------------------------------------------------------------------- User Interface Components
WithEvents Form1 As Form
WithEvents Timer1 As Timer = New Timer
Dim objThread As Threading.Thread
Dim Label1 As Label = New Label
Dim Label2 As Label = New Label
Dim Label3 As Label = New Label
Dim Label4 As Label = New Label
WithEvents TrackBar1 As TrackBar = New TrackBar

Sub InitializeComponents()
	Application.EnableVisualStyles
	form1.SuspendLayout
	' ------------------------------------------------------------------ Form
	Form1.Text = "Theremino Script - Geiger Simulator"
	Form1.ClientSize = New System.Drawing.Size(640, 320)
	Form1.MinimumSize = Form1.Size
	Form1.StartPosition = FormStartPosition.CenterScreen
	' ------------------------------------------------------------------ Label 1
	Label1.BorderStyle = BorderStyle.Fixed3D
	Label1.Location = New Point(10, 10)
	Label1.Font = New Font("Arial", 12 )
	Label1.Size = New Size(620, 25)
	Label1.TextAlign = ContentAlignment.MiddleLeft
	Label1.Anchor = AnchorStyles.top Or AnchorStyles.Left Or AnchorStyles.right
	Form1.Controls.Add(Label1)
	' ------------------------------------------------------------------ Label 2
	Label2.BorderStyle = BorderStyle.Fixed3D
	Label2.Location = New Point(10, 40)
	Label2.Font = New Font("Arial", 12 )
	Label2.Size = New Size(620, 25)
	Label2.TextAlign = ContentAlignment.MiddleLeft
	Label2.Anchor = AnchorStyles.top Or AnchorStyles.Left Or AnchorStyles.right
	Form1.Controls.Add(Label2)	
	' ------------------------------------------------------------------ Label 3
	Label3.BorderStyle = BorderStyle.Fixed3D
	Label3.Location = New Point(10, 70)
	Label3.Font = New Font("Arial", 12 )
	Label3.Size = New Size(620, 25)
	Label3.TextAlign = ContentAlignment.MiddleLeft
	Label3.Anchor = AnchorStyles.top Or AnchorStyles.Left Or AnchorStyles.right
	Form1.Controls.Add(Label3)	
	' ------------------------------------------------------------------ Label 4
	Label4.BorderStyle = BorderStyle.Fixed3D
	Label4.Location = New Point(10, 100)
	Label4.Font = New Font("Arial", 12 )
	Label4.Size = New Size(620, 25)
	Label4.TextAlign = ContentAlignment.MiddleLeft
	Label4.Anchor = AnchorStyles.top Or AnchorStyles.Left Or AnchorStyles.right
	Form1.Controls.Add(Label4)
	' ------------------------------------------------------------------ TrackBar 1
	TrackBar1.Location = New Point(10, 150)
	TrackBar1.Size = New Size(45, 150)
	TrackBar1.Orientation = Orientation.Vertical
	TrackBar1.TickStyle = TickStyle.Both
	TrackBar1.Minimum = 0
	TrackBar1.Maximum = 14
	TrackBar1.TickFrequency = 1
	TrackBar1.LargeChange = 1
	Form1.Controls.Add(TrackBar1)
	' ------------------------------------------------------------------ Timer 1
	Timer1.Interval = 50
	Timer1.Start
	' ------------------------------------------------------------------ 
	Form1.ResumeLayout
	Form1.PerformLayout()
	Form1.show
End Sub

