' ================================================================================================
' - ENGLISH - LOG CONVERTER
' ================================================================================================
'  This script reads "Sint Wind PI" LOG files and extracts meteo data lines in a CSV file.
' ------------------------------------------------------------------------------------------------

' ================================================================================================
' - ITALIANO - LOG CONVERTER
' ================================================================================================
'  Questo script legge un file LOG di "Sint Wind PI" e estrae i dati meteo in un file CSV.
' ------------------------------------------------------------------------------------------------

Dim ProgramName As String = "LOG CONVERTER"

Dim ValidData As String = "Dir: Spd: Gst: Tout: Hout: P: Rtot: RDay: R24h:"

Dim ofd As OpenFileDialog 
Dim sr As IO.StreamReader 
Dim sw As IO.StreamWriter 
Dim InString As String 
Dim OutString As String
Dim InCounter As Int32
Dim OutCounter As Int32

Sub OpenAndConvertFile
	' --------------------------------------------------------- open the file
	ofd = New OpenFileDialog
	ofd.Filter = "log files (*.log)|*.log|All files (*.*)|*.*"
	ofd.ShowDialog 
	If Not io.file.exists(ofd.FileName) Then Return
	sr = New IO.StreamReader(ofd.FileName)
	sw = New IO.StreamWriter(ofd.FileName.ToLower.Replace(".log",".csv"))
	InCounter = 0
	OutCounter = 0
	Label1.text = "Working"
	' ---------------------------------------------------------
	While Not sr.EndOfStream
		' ----------------------------------------------------- read a file line
		InString = sr.ReadLine
		InCounter += 1
		' ----------------------------------------------------- decode a file line
		OutString = DecodeFileLine(InString)
		' ----------------------------------------------------- valid line ?
		If OutString <> "" Then 
			' ------------------------------------------------- write line to out file
			sw.WriteLine(OutString)
			OutCounter += 1
		End If
	End While
	sr.close
	sw.close
	' --------------------------------------------------------- final message
	Label1.text = "In file: " & _
				InCounter.ToString & " lines" & vbcrlf & _
				"Out file: " & OutCounter.ToString & " lines"
End Sub


Function DecodeFileLine(ByVal line As String) As String
	' ---------------------------------------------------------- split the line in sections
	Dim Sections() As String = Split(line, " ")
	' ---------------------------------------------------------- test the first data section
	If Sections.Length > 1 AndAlso Sections(1) = "Dir:" Then
		' ------------------------------------------------------ convert date
		line = Sections(0)
		line = Mid(line, 2, 16)
		line = line.Replace("-", " ")
		' ------------------------------------------------------ extract valid meteo fields
		For i As Int32 = 1 To Sections.Length - 2
			If ValidData.Contains(Sections(i)) Then
				line += "; "
				line += Sections(i + 1) 
				i += 2
			End If
		Next
		line = line.Replace(".", ",")
		Return line
	Else
		Return ""
	End If
End Function




' ---------------------------------------------------------------------- 
'  User Interface Components
' ---------------------------------------------------------------------- 
WithEvents Form1 As Form
Dim Label1 As Label = New Label
Dim WithEvents ButtonOpenFile As System.Windows.Forms.Button = New System.Windows.Forms.Button

Sub ButtonOpenFile_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ButtonOpenFile.Click
	OpenAndConvertFile
End Sub

Sub Main(ByVal s() As String)
	If Form1 IsNot Nothing Then Return
	Form1 = New Form
	InitializeComponents
	While Form1.Visible
		Application.DoEvents
		Threading.Thread.Sleep(10)
	End While
End Sub

Sub InitializeComponents()
	Application.EnableVisualStyles
	form1.SuspendLayout
	' ------------------------------------------------------------------ Form
	Form1.Text = "Theremino Script - " & ProgramName
	Form1.FormBorderStyle = FormBorderStyle.Fixed3D
	Form1.ClientSize = New System.Drawing.Size(370, 60)
	Form1.MinimumSize = Form1.Size
	Form1.StartPosition = FormStartPosition.CenterScreen
	' ------------------------------------------------------------------ ButtonOpenFile
	ButtonOpenFile.Location = New Point(120, 0)
	ButtonOpenFile.Size = New Size(120, 23)
	ButtonOpenFile.Text = "File Open"
	ButtonOpenFile.UseVisualStyleBackColor = True
	Form1.Controls.Add(ButtonOpenFile)
	' ------------------------------------------------------------------ Label
	Label1.Location = New Point(4, 28)
	Label1.Width = Form1.ClientSize.Width - Label1.Left - 4
	Label1.Height = Form1.ClientSize.Height - Label1.Top - 4
	Label1.BackColor = Color.AliceBlue
	Label1.ForeColor = Color.Black
	Label1.Font = New Font("Tahoma", 10, FontStyle.Regular)
	Label1.TextAlign = ContentAlignment.MiddleLeft
	Form1.Controls.Add(Label1)
	' ------------------------------------------------------------------ 
	Form1.ResumeLayout
	Form1.PerformLayout()
	Form1.show
End Sub

