' ================================================================================================
' - ENGLISH - THEREMINO MANAGEMENT (Theremino System SLOTS operations)
' ================================================================================================
'  This script attempts to simplify the management of simple operations
'  Like "light a led", "wait a second", "turn off the led", "repeat three times"
'  The DotNet complexities are hidden as much as possible.
' ------------------------------------------------------------------------------------------------
'  The simplified approach is valid only for simple programs TEST.
'  If the project extends over a few pages it is better to use Microsoft's Visual Studio
'  that is the elder brother of this program.
'  The Express version of Visual Studio is free and works very well.
' ------------------------------------------------------------------------------------------------
'  Per accedere agli SLOTS usare le funzioni: "MemoryMappedFile_Init", "ReadSlot" and "WriteSlot"
' ------------------------------------------------------------------------------------------------
'  It's good to know that every RUN the program is automatically saved.
' ------------------------------------------------------------------------------------------------
'  If something goes wrong, save, close and reopen.
' ------------------------------------------------------------------------------------------------

' ================================================================================================
' - ITALIANO - THEREMINO MANAGEMENT (Operazioni sugli slot del sistema Theremino)
' ================================================================================================
'  Questo script cerca di semplificare al massimo la gestione di semplici operazioni
'  del tipo "accendo un led", "attendo un secondo", "spengo il led", "ripeto tre volte"
'  Le complessita' di DotNet vengono nascoste per quanto possibile.
' ------------------------------------------------------------------------------------------------
'  L'approccio semplificato e' valido solo per semplici programmi di TEST.
'  Se il progetto si allunga oltre qualche pagina e' meglio usare Visual Studio di Microsoft
'   che e' il fratello maggiore di questo programma.
'  La versione Express di Visual Studio e' gratuita e funziona molto bene.
' ------------------------------------------------------------------------------------------------
'  To access Theremino SLOTS use: "MemoryMappedFile_Init", "ReadSlot" and "WriteSlot" functions
' ------------------------------------------------------------------------------------------------
'  E' bene sapere che ad ogni RUN il programma viene salvato automaticamente.
' ------------------------------------------------------------------------------------------------
'  Se qualcosa va storto, salvare, chiudere e riaprire.
' ================================================================================================



' - ENGLISH --------------------------------------------------------------------------------------
'  Global variables, used by all functions, must be dimensioned here/ 
'  Normally for the SLOT values variables the type "Single" is used
'  For integer numbers use variables of type "Int32"
'  Examples:
'  Dim gv1 as Single
'  Dim gv2 as Single
'  Dim gi1 as Int32
'  Dim gi2 as Int32
'  The initial "g" stands for "global" is used for shared variables, usable by all functions
'  The "v" stands for "value", usually numerical values of the slots, usually of type "Single"
'  The "i" stands for "integer", used for counters and indices in matrices
'  Instead of the final number it is a good idea to use meaningful names (eg. GiLoopCounter)
' ------------------------------------------------------------------------------------------------

' - ITALIAN --------------------------------------------------------------------------------------
'  Dimensionare qui le variabili globali che devono essere usate da tutte le funzioni
'  Normalmente per i valori da leggere scrivere negli SLOT si usano variabili di tipo "Single"
'  Per i numeri interi e' bene usare variabili di tipo "Int32"
'  Esempi:
'  Dim gv1 as Single
'  Dim gv2 as Single
'  Dim gi1 as Int32
'  Dim gi2 as Int32
'  La "g" iniziale sta per "globale" cioe' una variabile condivisa e usabile da tutte le funzioni
'  La "v" sta per "value", normalmente valori numerici degli slot, normalmente di tipo "Single"
'  La "i" sta per "intera", si usano spesso nei "For", come contatori e come indici nelle matrici
'  Al posto del numero finale e' una buona idea usare nomi significativi (ad es. giLoopCounter)
' ------------------------------------------------------------------------------------------------
Dim gv1 as Single


' ------------------------------------------------------------------------------------------------
'  - ENG - This function is executed only at the program start - Do here all the initializatons
' ------------------------------------------------------------------------------------------------
'  - ITA - Questa funzione viene esguita solo alla partenza - Fare qui tutte le inizializzazioni
' ------------------------------------------------------------------------------------------------
Sub Initialize
	
End Sub


' - ENGLISH --------------------------------------------------------------------------------------
'  By default this function is called 20 times per second
'  It is possible to change the calling time with: "Timer1.Interval = xxx" (xxx is time in ms)
'  All the main window controls updates must be implemented here
'  All the frequent slot updates must be done here
'  No PAUSES here, only fast executions.
' ------------------------------------------------------------------------------------------------

' - ITALIAN --------------------------------------------------------------------------------------
'  Per default funzione viene chiamata 20 volte al secondo
'  E' possibile modificare il tempo di chiamate con: "Timer1.Interval = xxx" (xxx = tempo in ms)
'  Tutte le visualizzazioni dei dati nei controlli della finestra principale devono avvenire qui
'  Se si devono leggere o scrivere i valori degli slot con grande frequenza e' bene farlo qui
'  Niente PAUSE qui dentro, solo esecuzioni piu' rapide possibile
' ------------------------------------------------------------------------------------------------
Sub FastExecution
	gv1 = ReadSlot(5)
	Label1.Text = gV1.ToString("0") 
	If ReadSlot(5) > 500 Then
		WriteSlot(15, gv1)
		Label2.Text = "Writing to slot 15"
	else
		Label2.Text = ""
	End If
End Sub

' - ENGLISH --------------------------------------------------------------------------------------
'  Here you can perform any kind of slow operation.
'  You can do PAUSE long as you like, cycles LOOP, FOR, WHILE, etc. ...
'  Inside every tight loop must be at least a break from 1 millisecond
'   otherwise the CPU starts to consume 100% and heats.
' ------------------------------------------------------------------------------------------------
'  In this function you can read and write SLOTS
'  But it is well NOT to write repeatedly in the main window
'  data updates should be done in the "FastExecution" function
' ------------------------------------------------------------------------------------------------
'  If you come to the "End Sub", then the function starts immediately again
' ------------------------------------------------------------------------------------------------

' - ITALIANO -------------------------------------------------------------------------------------
'  Qui si puo' eseguire ogni genere di operazione lenta
'  Si possono fare PAUSE lunghe a piacere, cicli LOOP, FOR, WHILE etc...
'  Dentro ad ogni ciclo stretto deve esserci almeno una pausa da 1 MilliSecondo
'   altrimenti la CPU si mette a consumare il 100% e scalda da bestia
' ------------------------------------------------------------------------------------------------
'  In questa funzione si possono leggere e scrivere gli SLOT
'  Ma e' bene NON scrivere ripetutamente nella finestra principale (TEXT BOX)
'  gli aggiornamenti dei dati vanno fatte nella funzione "FastExecution"
' ------------------------------------------------------------------------------------------------
'  Se si arriva all' "End Sub" allora la funzione ricomincia immediatamente da capo
' ------------------------------------------------------------------------------------------------
Sub PausableExecution()
	Label4.Text = "Eseguo la prima parte"
	' ---------------------------------- accendo e spengo il led sullo slot 1 per 3 volte		  
	For i1 As Int32 = 1 to 3
		Label3.Text = i1.ToString
		PausaMS(200)
		WriteSlot(1, 1000)
		PausaMS(200)
		WriteSlot(1, 0)
	Next
	Label4.Text = "Attendo 2 secondi"
	PausaMS(2000)
	Label4.Text = "Accendo progressivamente il led (usare una uscita Pwm16 - logaritmica)"
	' ---------------------------------- accendo progressivamente il led 1 da 0 (spento) a 1000 (massimo)
	'                                    in 100 passi distanti 100 milli secondi
	For i2 As Int32 = 0 to 100
		WriteSlot(1, i2 * 10)
		PausaMS(100)
		Label3.Text = i2.ToString
	Next
	Label3.Text = ""
	' ---------------------------------- segnalo di aver finito
	'Beep()
	Label4.Text = "Fine ciclo - Attendo 2 secondi"
	PausaMS(2000)
End Sub


' ------------------------------------------------------------------------------------------------
'  Queste funzioni vengono chiamate dagli eventi causati dai controlli utente
' ------------------------------------------------------------------------------------------------
Sub TrackBar1_Scroll(value As Single)
	WriteSlot(16, value)
	Label3.Text = "E' stato mosso il cursore, il suo valore e': " & value.ToString("0")
End Sub



' ================================================================================================
'  - ENG - THE SIMPLIFIED MANAGEMENT ENDS HER - MODIFY THE FOLLOWING CAREFULLY
' ------------------------------------------------------------------------------------------------
'  - ITA - LA GESTIONE SEMPLIFICATA FINISCE QUI - MODIFICARE LA PARTE SEGUENTE CON ATTENZIONE
' ================================================================================================


Sub Main(ByVal s() As String)
	if Form1 IsNot Nothing then return
	Form1 = New form
	InitializeComponents
	MemoryMappedFile_Init
	StartThread
	While Form1.Visible
		Application.DoEvents
		Threading.Thread.Sleep(10)
	End While
End Sub

Sub Form1_Closing(sender As Object, e As CancelEventArgs) Handles Form1.Closing
	Timer1.Stop
	objThread.Abort
End Sub

Sub StartThread()
	' --------------------------------------------------------- asynchronously start the Thread 
	objThread = New Threading.Thread(AddressOf ExecuteThread)
	' --------------------------------------------------------- make the thread as background thread.
	objThread.IsBackground = True
	' --------------------------------------------------------- set the Priority of the thread.
	objThread.Priority = Threading.ThreadPriority.AboveNormal
	' --------------------------------------------------------- start the thread.
	objThread.Start()
End Sub

Sub ExecuteThread
	Do
		PausableExecution
		PausaMS(1)
	loop until false
End Sub

Sub PausaMS(Byval mS As Int32)
	Threading.Thread.Sleep(mS)
End Sub

Sub Timer1_Tick(ByVal sender As Object, ByVal e As EventArgs) Handles Timer1.Tick
	FastExecution
End Sub

Sub TrackBar1_Scroll(ByVal sender As Object, ByVal e As EventArgs) Handles TrackBar1.Scroll
	TrackBar1_Scroll(TrackBar1.Value)
End Sub

' ---------------------------------------------------------------------- User Interface Components
WithEvents Form1 As Form
WithEvents Timer1 As Timer = New Timer
Dim objThread As Threading.Thread
Dim Label1 As Label = new Label
Dim Label2 As Label = new Label
Dim Label3 As Label = new Label
Dim Label4 As Label = new Label
WithEvents TrackBar1 As TrackBar = new TrackBar

Sub InitializeComponents()
	Application.EnableVisualStyles
	form1.SuspendLayout
	' ------------------------------------------------------------------ Form
	Form1.Text = "Theremino Script - Test 1"
	Form1.ClientSize = New System.Drawing.Size(640, 320)
	Form1.MinimumSize = Form1.Size
	Form1.StartPosition = FormStartPosition.CenterScreen
	' ------------------------------------------------------------------ Label 1
	Label1.BorderStyle = BorderStyle.Fixed3D
	Label1.Location = New Point(10, 10)
	Label1.Font = New Font("Arial", 12 )
	Label1.Size = New Size(620, 21)
	Label1.Anchor = AnchorStyles.top or AnchorStyles.Left or AnchorStyles.right
	Form1.Controls.Add(Label1)
	' ------------------------------------------------------------------ Label 2
	Label2.BorderStyle = BorderStyle.Fixed3D
	Label2.Location = New Point(10, 40)
	Label2.Font = New Font("Arial", 12 )
	Label2.Size = New Size(620, 21)
	Label2.Anchor = AnchorStyles.top or AnchorStyles.Left or AnchorStyles.right
	Form1.Controls.Add(Label2)	
	' ------------------------------------------------------------------ Label 3
	Label3.BorderStyle = BorderStyle.Fixed3D
	Label3.Location = New Point(10, 70)
	Label3.Font = New Font("Arial", 12 )
	Label3.Size = New Size(620, 21)
	Label3.Anchor = AnchorStyles.top or AnchorStyles.Left or AnchorStyles.right
	Form1.Controls.Add(Label3)	
	' ------------------------------------------------------------------ Label 4
	Label4.BorderStyle = BorderStyle.Fixed3D
	Label4.Location = New Point(10, 100)
	Label4.Font = New Font("Arial", 12 )
	Label4.Size = New Size(620, 21)
	Label4.Anchor = AnchorStyles.top or AnchorStyles.Left or AnchorStyles.right
	Form1.Controls.Add(Label4)
	' ------------------------------------------------------------------ TrackBar 1
	TrackBar1.Location = New Point(10, 130)
	TrackBar1.Size = New Size(45, 149)
	TrackBar1.Orientation = Orientation.Vertical
	TrackBar1.TickStyle = TickStyle.Both
	TrackBar1.Maximum = 1000
	TrackBar1.TickFrequency = 100
	TrackBar1.LargeChange = 10
	Form1.Controls.Add(TrackBar1)
	' ------------------------------------------------------------------ Timer 1
	Timer1.Interval = 50
	Timer1.Start
	' ------------------------------------------------------------------ 
	Form1.ResumeLayout
	Form1.PerformLayout()
	Form1.show
End Sub

